import os
import random
import sys
import threading
import unittest
from timeit import timeit

import helpers
from helpers import TestCase, get_method_names, get_methods
from screen_brightness_control.exceptions import NoValidDisplayError

sys.path.insert(0, os.path.abspath('./'))
import screen_brightness_control as sbc  # noqa: E402


class TestGetBrightness(TestCase):
    def test_normal(self):
        brightness = sbc.get_brightness()
        # check it is the right type and within
        # the max and min values
        self.assertBrightnessValid(brightness)

    def test_display_kwarg(self):
        for monitor in sbc.list_monitors():
            brightness = sbc.get_brightness(display=monitor)
            self.assertBrightnessValid(brightness, target_length=1)

    def test_method_kwarg(self):
        for method in get_method_names():
            try:
                value = sbc.get_brightness(method=method)
                self.assertBrightnessValid(value, target_length=len(sbc.list_monitors_info(method=method)))
            except NoValidDisplayError:
                pass

    def test_abnormal(self):
        self.assertRaises(ValueError, sbc.get_brightness, method='non-existant method')
        self.assertRaises(NoValidDisplayError, sbc.get_brightness, display='non-existant display')
        # test wrong types
        self.assertRaises(TypeError, sbc.get_brightness, method=0.0)
        self.assertRaises(TypeError, sbc.get_brightness, display=0.0)


class TestSetBrightness(TestCase):
    def test_normal(self):
        for value in (0, 10, 21, 37, 43, 50, 90, 100):
            brightness = sbc.set_brightness(value, force=True, no_return=False)
            self.assertIsNone(sbc.set_brightness(value, force=True))
            self.assertBrightnessValid(brightness)
            for index, i in enumerate(brightness):
                # use almost equal because some laptops cannot display all values 0 to 100
                self.assertBrightnessEqual(value, i, index)

    def test_increment_values(self):
        kw = {'display': 0} if sbc.list_monitors_info() else {}
        kw['no_return'] = False

        self.assertBrightnessEqual(sbc.set_brightness('50', **kw), [50], 0)
        self.assertBrightnessEqual(sbc.set_brightness('10.0', **kw), [10], 0)
        self.assertBrightnessEqual(sbc.set_brightness('+40', **kw), [50], 0)
        self.assertBrightnessEqual(sbc.set_brightness('-20', **kw), [30], 0)
        self.assertBrightnessEqual(sbc.set_brightness('+500', **kw), [100], 0)

        # test that all displays are affected equally
        for _ in sbc.list_monitors():
            sbc.set_brightness(random.randint(30, 70))

        old = sbc.get_brightness()
        new = sbc.set_brightness('-25', no_return=False)
        for i, v in enumerate(old):
            self.assertBrightnessEqual(max(0, v - 25), new[i], i)

    def test_display_kwarg(self):
        for index, monitor in enumerate(sbc.list_monitors()):
            brightness = sbc.set_brightness(90, display=monitor, no_return=False)
            self.assertBrightnessValid(brightness, target_length=1)
            self.assertBrightnessEqual(90, brightness[0], index)

    def test_method_kwarg(self):
        for method in get_method_names():
            try:
                value = sbc.set_brightness(90, method=method, no_return=False)
                self.assertBrightnessValid(value, target_length=len(sbc.list_monitors_info(method=method)))
            except NoValidDisplayError:
                pass

    def test_abnormal(self):
        self.assertRaises(ValueError, sbc.set_brightness, 100, method='non-existant method')
        self.assertRaises(NoValidDisplayError, sbc.set_brightness, 100, display='non-existant display')
        # test wrong types
        self.assertRaises(TypeError, sbc.set_brightness, 100, method=0.0)
        self.assertRaises(TypeError, sbc.set_brightness, 100, display=0.0)


class TestFadeBrightness(TestCase):
    def test_normal(self):
        brightness = sbc.fade_brightness(90)
        self.assertBrightnessValid(brightness, target_length=len(sbc.list_monitors()))
        for index, i in enumerate(brightness):
            self.assertBrightnessEqual(90, i, index)

    def test_increment_values(self):
        with helpers.FakeMethodTest():
            self.assertBrightnessEqual(sbc.fade_brightness('90', display=0), [90], 0)
            self.assertBrightnessEqual(sbc.fade_brightness('80.0', display=0), [80], 0)
            self.assertBrightnessEqual(sbc.fade_brightness('+10', display=0), [90], 0)
            self.assertBrightnessEqual(sbc.fade_brightness('-10', display=0), [80], 0)
            self.assertBrightnessEqual(sbc.fade_brightness('+500', display=0), [100], 0)

            # test that all displays are affected equally
            for _ in sbc.list_monitors():
                sbc.set_brightness(random.randint(30, 70))

            old = sbc.get_brightness()
            new = sbc.fade_brightness('-25')
            for i, v in enumerate(old):
                self.assertBrightnessEqual(max(0, v - 25), new[i], i)

    def test_increment_kwarg(self):
        # smaller increment should take longer
        self.assertGreater(
            timeit(lambda: sbc.fade_brightness(80, start=100, increment=1), number=1),
            timeit(lambda: sbc.fade_brightness(80, start=100, increment=5), number=1)
        )

    def test_interval_kwarg(self):
        # longer intervals should take longer
        self.assertGreater(
            timeit(lambda: sbc.fade_brightness(80, start=100, interval=0.05), number=1),
            timeit(lambda: sbc.fade_brightness(80, start=100, interval=0.01), number=1)
        )

    def test_blocking_kwarg(self):
        threads = sbc.fade_brightness(90, blocking=False)
        self.assertIsInstance(threads, list)
        for i, thread in enumerate(threads):
            self.assertIsInstance(thread, threading.Thread)
            thread.join()
            self.assertBrightnessEqual(sbc.get_brightness(display=i)[0], 90, i)

    def test_display_kwarg(self):
        for index, monitor in enumerate(sbc.list_monitors()):
            brightness = sbc.fade_brightness(90, display=monitor)
            self.assertBrightnessValid(brightness, target_length=1)
            self.assertBrightnessEqual(brightness, [90], index)

    def test_method_kwarg(self):
        for method in get_method_names():
            try:
                value = sbc.fade_brightness(90, method=method)
                self.assertBrightnessValid(value, target_length=len(sbc.list_monitors_info(method=method)))
            except NoValidDisplayError:
                # likely no monitors of that method
                pass

    def test_abnormal(self):
        self.assertRaises(ValueError, sbc.fade_brightness, 100, method='non-existant method')
        self.assertRaises(NoValidDisplayError, sbc.fade_brightness, 100, display='non-existant display')
        # test wrong types
        self.assertRaises(TypeError, sbc.fade_brightness, 100, method=0.0)
        self.assertRaises(TypeError, sbc.fade_brightness, 100, display=0.0)


class TestListMonitorsInfo(TestCase):
    def test_normal(self):
        methods = get_methods()
        monitors = sbc.list_monitors_info()
        self.assertIsInstance(monitors, list)
        for monitor in monitors:
            self.assertIsInstance(monitor, dict)
            for key in ('name', 'model', 'serial', 'edid', 'manufacturer', 'manufacturer_id'):
                if monitor[key] is not None:
                    self.assertIsInstance(monitor[key], str)

            self.assertIsInstance(monitor['index'], int)
            self.assertIn(monitor['method'], methods)
            self.assertNotIn('unsupported', monitor)

    def test_method_kwarg(self):
        methods = get_methods()
        method_names = get_method_names()
        for name, method in zip(method_names, methods):
            for monitor in sbc.list_monitors_info(method=name):
                self.assertEqual(monitor['method'], method)

    def test_unsupported_kwarg(self):
        supported = sbc.list_monitors_info(allow_duplicates=True)
        incl_unsupported = sbc.list_monitors_info(allow_duplicates=True, unsupported=True)

        for display in supported:
            self.assertNotIn('unsupported', display)

        for display in incl_unsupported:
            if 'unsupported' in display:
                self.assertIsInstance(display['unsupported'], bool)
                display.pop('unsupported')
                self.assertIn(display, supported)


class TestListMonitors(TestCase):
    def test_normal(self):
        monitors = sbc.list_monitors()
        self.assertIsInstance(monitors, list)
        for monitor in monitors:
            self.assertIsInstance(monitor, str)


class TestGetMethods(TestCase):
    def test_normal(self):
        methods = sbc.get_methods()
        self.assertIsInstance(methods, dict)

        for name, method in methods.items():
            self.assertTrue(hasattr(method, 'get_brightness'))
            self.assertTrue(hasattr(method, 'set_brightness'))
            self.assertTrue(hasattr(method, 'get_display_info'))

            self.assertEqual(name.lower(), method.__name__.lower())

    def test_lookup(self):
        all_methods = sbc.get_methods()

        for name in all_methods.keys():
            self.assertEqual(sbc.get_methods(name), {name: all_methods[name]})

    def test_lookup_abnormal(self):
        self.assertRaises(ValueError, sbc.get_methods, 'not a method')


class TestDisplay(TestCase):
    primary: sbc.Display

    def setUp(self):
        super().setUp()
        self.primary = sbc.Display.from_dict(sbc.list_monitors_info()[0])

    def test_fade_brightness(self):
        # test normal
        brightness = self.primary.fade_brightness(75)
        self.assertIsInstance(brightness, int)
        self.assertTrue(0 <= brightness <= 100)
        self.assertBrightnessEqual(75, brightness, 0)

        # test increment values
        self.assertBrightnessEqual(self.primary.fade_brightness('60'), 60, 0)
        self.assertBrightnessEqual(self.primary.fade_brightness('70.0'), 70, 0)
        self.assertBrightnessEqual(self.primary.fade_brightness('+10'), 80, 0)
        self.assertBrightnessEqual(self.primary.fade_brightness('-10'), 70, 0)
        self.assertBrightnessEqual(
            self.primary.fade_brightness('+500'), 100, 0)

        # test increment kwarg
        # smaller increment should take longer
        self.assertGreater(
            timeit(lambda: self.primary.fade_brightness(
                90, start=100, increment=1), number=1),
            timeit(lambda: self.primary.fade_brightness(
                90, start=100, increment=2), number=1)
        )

    def test_from_dict(self):
        info = sbc.list_monitors_info()[1]
        info['something'] = 'extra'
        display = sbc.Display.from_dict(info)
        self.assertIsInstance(display, sbc.Display)
        # make sure that misc info wasn't assigned
        self.assertFalse(hasattr(display, 'something'))
        for prop in (
            'index', 'method', 'edid', 'manufacturer',
            'manufacturer_id', 'model', 'name', 'serial'
        ):
            self.assertIs(info[prop], getattr(display, prop))

    def test_get_brightness(self):
        brightness = self.primary.get_brightness()
        self.assertIsInstance(brightness, int)
        self.assertTrue(0 <= brightness <= 100)

    def test_get_identifier(self):
        identifier = self.primary.get_identifier()
        self.assertIsInstance(identifier, tuple)
        self.assertIn(identifier[0], ('edid', 'serial', 'name', 'index'))
        self.assertIsNotNone(identifier[1])
        if identifier[0] == 'index':
            self.assertIsInstance(identifier[1], int)
        else:
            self.assertIsInstance(identifier[1], str)

    def test_is_active(self):
        self.assertIsInstance(self.primary.is_active(), bool)

    def test_set_brightness(self):
        # test normal
        for value in (0, 10, 21, 37, 43, 50, 90, 100):
            ret_val = self.primary.set_brightness(value)
            self.assertIsNone(ret_val)


class TestMonitor(TestCase):
    def test_normal(self):
        primary = sbc.list_monitors_info()[0]
        monitor = sbc.Monitor(0)
        self.assertDictEqual(monitor.get_info(), primary)
        monitor_vars = {k: v for k, v in vars(sbc.Monitor(primary['name'])).items() if not k.startswith('_')}
        self.assertDictEqual(monitor_vars, primary)

    def test_get_identifier(self):
        monitor = sbc.Monitor(0)
        identifier = monitor.get_identifier()
        self.assertIsInstance(identifier, tuple)
        self.assertIn(identifier[0], ('edid', 'serial', 'name', 'index'))
        self.assertIsNotNone(identifier[1])
        if identifier[0] == 'index':
            self.assertIsInstance(identifier[1], int)
        else:
            self.assertIsInstance(identifier[1], str)

    def test_set_brightness(self):
        monitor = sbc.Monitor(0)

        # test normal
        for value in (0, 10, 21, 37, 43, 50, 90, 100):
            brightness = monitor.set_brightness(value, no_return=False)
            # check it is the right type and within
            # the max and min values
            self.assertIsInstance(brightness, int)
            self.assertTrue(0 <= brightness <= 100)
            # use almost equal because some laptops cannot display all values 0 to 100
            self.assertBrightnessEqual(value, brightness, 0)

        self.assertIsNone(monitor.set_brightness(100))
        self.assertIsNone(monitor.set_brightness(100, no_return=True))

    def test_get_brightness(self):
        monitor = sbc.Monitor(0)
        brightness = monitor.get_brightness()
        self.assertIsInstance(brightness, int)
        self.assertTrue(0 <= brightness <= 100)

    def test_fade_brightness(self):
        monitor = sbc.Monitor(0)

        # test normal
        brightness = monitor.fade_brightness(75)
        self.assertIsInstance(brightness, int)
        self.assertTrue(0 <= brightness <= 100)
        self.assertBrightnessEqual(75, brightness, 0)

        # test increment values
        self.assertBrightnessEqual(monitor.fade_brightness('60'), 60, 0)
        self.assertBrightnessEqual(monitor.fade_brightness('70.0'), 70, 0)
        self.assertBrightnessEqual(monitor.fade_brightness('+10'), 80, 0)
        self.assertBrightnessEqual(monitor.fade_brightness('-10'), 70, 0)
        self.assertBrightnessEqual(monitor.fade_brightness('+500'), 100, 0)

        # test increment kwarg
        # smaller increment should take longer
        self.assertGreater(
            timeit(lambda: monitor.fade_brightness(90, start=100, increment=1), number=1),
            timeit(lambda: monitor.fade_brightness(90, start=100, increment=2), number=1)
        )

        # test blocking kwarg
        thread = monitor.fade_brightness(100, blocking=False)
        self.assertIsInstance(thread, threading.Thread)
        thread.join()
        self.assertBrightnessEqual(monitor.get_brightness(), 100, 0)

    def test_get_info(self):
        monitor = sbc.Monitor(0)
        self.assertIsInstance(monitor.get_info(), dict)
        monitor.name = 'overwrite name'
        self.assertEqual(monitor.get_info(refresh=False)['name'], 'overwrite name')
        self.assertNotEqual(monitor.get_info(refresh=True)['name'], 'overwrite name')

    def test_is_active(self):
        self.assertIsInstance(sbc.Monitor(0).is_active(), bool)


class TestFilterMonitors(TestCase):
    def test_normal(self):
        monitors = sbc.list_monitors_info()
        filtered = sbc.filter_monitors()
        self.assertEqual(monitors, filtered)
        self.assertIsInstance(filtered, list)
        for monitor in filtered:
            self.assertIsInstance(monitor, dict)
            for key in ('name', 'model', 'serial', 'edid', 'manufacturer', 'manufacturer_id'):
                if monitor[key] is not None:
                    self.assertIsInstance(monitor[key], str)

    def test_display_kwarg(self):
        monitor = sbc.list_monitors_info()[0]
        self.assertEqual(sbc.filter_monitors(display=0)[0], monitor)
        for identifier in ('edid', 'serial', 'name', 'model'):
            if monitor[identifier] is not None:
                filtered = sbc.filter_monitors(display=monitor[identifier])
                if len(filtered) == 1:
                    self.assertEqual(monitor, filtered[0])
                else:
                    self.assertIn(monitor, filtered)

        self.assertRaises(TypeError, sbc.filter_monitors, display=0.0)
        self.assertRaises(TypeError, sbc.filter_monitors, display=[])
        self.assertRaises(
            NoValidDisplayError, sbc.filter_monitors, display="doesn't exist"
        )

    def test_haystack_kwarg(self):
        monitors = sbc.list_monitors_info()
        haystack = monitors + [monitors[0].copy()]
        haystack[-1]['name'] = "Monitor name that doesn't exist"

        # check that it filters out our duplicate entry
        self.assertEqual(sbc.filter_monitors(haystack=haystack), monitors)
        # now check that we can search for the duplicate entry
        self.assertEqual(
            sbc.filter_monitors(
                display="Monitor name that doesn't exist",
                haystack=haystack
            )[0], haystack[-1]
        )

    def test_method_kwarg(self):
        methods = get_methods()
        method_names = get_method_names()

        for method, name in zip(methods, method_names):
            try:
                for monitor in sbc.filter_monitors(method=name):
                    self.assertEqual(monitor['method'], method)
            except NoValidDisplayError:
                # if we don't have monitors of a certain method then pass
                pass

        self.assertRaises(ValueError, sbc.filter_monitors, method='not a method')
        self.assertRaises(ValueError, sbc.filter_monitors, method='not a method', haystack=[])

    def test_include_kwarg(self):
        monitors = sbc.list_monitors_info()
        haystack = monitors + [monitors[0].copy()]
        haystack[-1]['extra field'] = 'Extra info'

        self.assertEqual(
            sbc.filter_monitors(
                display='Extra info',
                haystack=haystack,
                include=['extra field']
            )[0], haystack[-1]
        )
        self.assertRaises(
            NoValidDisplayError,
            sbc.filter_monitors,
            display='Extra info',
            haystack=haystack
        )


if __name__ == '__main__':
    if '--synthetic' in sys.argv:
        sys.argv.remove('--synthetic')
        helpers.TEST_FAST = True
    else:
        helpers.TEST_FAST = False

    unittest.main()



import fcntl
import functools
import glob
import logging
import operator
import os
import re
import time
from typing import List, Optional, Tuple

from . import filter_monitors, get_methods
from .exceptions import I2CValidationError, NoValidDisplayError, format_exc
from .helpers import (EDID, BrightnessMethod, BrightnessMethodAdv, __Cache,
                      _monitor_brand_lookup, check_output)
from .types import DisplayIdentifier, IntPercentage

__cache__ = __Cache()
logger = logging.getLogger(__name__)


class SysFiles(BrightnessMethod):
    '''
    A way of getting display information and adjusting the brightness
    that does not rely on any 3rd party software.

    This class works with displays that show up in the `/sys/class/backlight`
    directory (so usually laptop displays).

    To set the brightness, your user will need write permissions for
    `/sys/class/backlight/*/brightness` or you will need to run the program
    as root.
    '''
    logger = logger.getChild('SysFiles')

    @classmethod
    def get_display_info(cls, display: Optional[DisplayIdentifier] = None) -> List[dict]:
        subsystems = set()
        for folder in os.listdir('/sys/class/backlight'):
            if os.path.isdir(f'/sys/class/backlight/{folder}/subsystem'):
                subsystems.add(tuple(os.listdir(f'/sys/class/backlight/{folder}/subsystem')))

        all_displays = {}
        index = 0

        for subsystem in subsystems:

            device = {
                'name': subsystem[0],
                'path': f'/sys/class/backlight/{subsystem[0]}',
                'method': cls,
                'index': index,
                'model': None,
                'serial': None,
                'manufacturer': None,
                'manufacturer_id': None,
                'edid': None,
                'scale': None
            }

            for folder in subsystem:
                # subsystems like intel_backlight usually have an acpi_video0
                # counterpart, which we don't want so lets find the 'best' candidate
                try:
                    with open(os.path.join(f'/sys/class/backlight/{folder}/max_brightness')) as f:
                        scale = int(f.read().rstrip(' \n')) / 100

                    # use the display with the highest resolution scale
                    if device['scale'] is None or scale > device['scale']:
                        device['name'] = folder
                        device['path'] = f'/sys/class/backlight/{folder}'
                        device['scale'] = scale
                except (FileNotFoundError, TypeError) as e:
                    cls.logger.error(
                        f'error getting highest resolution scale for {folder}'
                        f' - {format_exc(e)}'
                    )
                    continue

            if os.path.isfile('%s/device/edid' % device['path']):
                device['edid'] = EDID.hexdump('%s/device/edid' % device['path'])

                for key, value in zip(
                    ('manufacturer_id', 'manufacturer', 'model', 'name', 'serial'),
                    EDID.parse(device['edid'])
                ):
                    if value is None:
                        continue
                    device[key] = value

            all_displays[device['edid']] = device
            index += 1

        all_displays = list(all_displays.values())
        if display is not None:
            all_displays = filter_monitors(
                display=display, haystack=all_displays, include=['path'])
        return all_displays

    @classmethod
    def get_brightness(cls, display: Optional[int] = None) -> List[IntPercentage]:
        info = cls.get_display_info()
        if display is not None:
            info = [info[display]]

        results = []
        for device in info:
            with open(os.path.join(device['path'], 'brightness'), 'r') as f:
                brightness = int(f.read().rstrip('\n'))
            results.append(int(brightness / device['scale']))

        return results

    @classmethod
    def set_brightness(cls, value: IntPercentage, display: Optional[int] = None):
        info = cls.get_display_info()
        if display is not None:
            info = [info[display]]

        for device in info:
            with open(os.path.join(device['path'], 'brightness'), 'w') as f:
                f.write(str(int(value * device['scale'])))


class I2C(BrightnessMethod):
    '''
    In the same spirit as `SysFiles`, this class serves as a way of getting
    display information and adjusting the brightness without relying on any
    3rd party software.

    Usage of this class requires read and write permission for `/dev/i2c-*`.

    This class works over the I2C bus, primarily with desktop monitors as I
    haven't tested any e-DP displays yet.

    Massive thanks to [siemer](https://github.com/siemer) for
    his work on the [ddcci.py](https://github.com/siemer/ddcci) project,
    which served as a my main reference for this.

    References:
        * [ddcci.py](https://github.com/siemer/ddcci)
        * [DDCCI Spec](https://milek7.pl/ddcbacklight/ddcci.pdf)
    '''
    logger = logger.getChild('I2C')

    # vcp commands
    GET_VCP_CMD = 0x01
    '''VCP command to get the value of a feature (eg: brightness)'''
    GET_VCP_REPLY = 0x02
    '''VCP feature reply op code'''
    SET_VCP_CMD = 0x03
    '''VCP command to set the value of a feature (eg: brightness)'''

    # addresses
    DDCCI_ADDR = 0x37
    '''DDC packets are transmittred using this I2C address'''
    HOST_ADDR_R = 0x50
    '''Packet source address (the computer) when reading data'''
    HOST_ADDR_W = 0x51
    '''Packet source address (the computer) when writing data'''
    DESTINATION_ADDR_W = 0x6e
    '''Packet destination address (the monitor) when writing data'''
    I2C_SLAVE = 0x0703
    '''The I2C slave address'''

    # timings
    WAIT_TIME = 0.05
    '''How long to wait between I2C commands'''

    _max_brightness_cache: dict = {}

    class I2CDevice():
        '''
        Class to read and write data to an I2C bus,
        based on the `I2CDev` class from [ddcci.py](https://github.com/siemer/ddcci)
        '''

        def __init__(self, fname: str, slave_addr: int):
            '''
            Args:
                fname: the I2C path, eg: `/dev/i2c-2`
                slave_addr: not entirely sure what this is meant to be
            '''
            self.device = os.open(fname, os.O_RDWR)
            # I2C_SLAVE address setup
            fcntl.ioctl(self.device, I2C.I2C_SLAVE, slave_addr)

        def read(self, length: int) -> bytes:
            '''
            Read a certain number of bytes from the I2C bus

            Args:
                length: the number of bytes to read

            Returns:
                bytes
            '''
            return os.read(self.device, length)

        def write(self, data: bytes) -> int:
            '''
            Writes data to the I2C bus

            Args:
                data: the data to write

            Returns:
                The number of bytes written
            '''
            return os.write(self.device, data)

    class DDCInterface(I2CDevice):
        '''
        Class to send DDC (Display Data Channel) commands to an I2C device,
        based on the `Ddcci` and `Mccs` classes from [ddcci.py](https://github.com/siemer/ddcci)
        '''

        PROTOCOL_FLAG = 0x80

        def __init__(self, i2c_path: str):
            '''
            Args:
                i2c_path: the path to the I2C device, eg: `/dev/i2c-2`
            '''
            self.logger = logger.getChild(
                self.__class__.__name__).getChild(i2c_path)
            super().__init__(i2c_path, I2C.DDCCI_ADDR)

        def write(self, *args) -> int:
            '''
            Write some data to the I2C device.

            It is recommended to use `setvcp` to set VCP values on the DDC device
            instead of using this function directly.

            Args:
                *args: variable length list of arguments. This will be put
                    into a `bytearray` and wrapped up in various flags and
                    checksums before being written to the I2C device

            Returns:
                The number of bytes that were written
            '''
            time.sleep(I2C.WAIT_TIME)

            ba = bytearray(args)
            ba.insert(0, len(ba) | self.PROTOCOL_FLAG)  # add length info
            ba.insert(0, I2C.HOST_ADDR_W)  # insert source address
            ba.append(functools.reduce(operator.xor, ba,
                      I2C.DESTINATION_ADDR_W))  # checksum

            return super().write(ba)

        def setvcp(self, vcp_code: int, value: int) -> int:
            '''
            Set a VCP value on the device

            Args:
                vcp_code: the VCP command to send, eg: `0x10` is brightness
                value: what to set the value to

            Returns:
                The number of bytes written to the device
            '''
            return self.write(I2C.SET_VCP_CMD, vcp_code, *value.to_bytes(2, 'big'))

        def read(self, amount: int) -> bytes:
            '''
            Reads data from the DDC device.

            It is recommended to use `getvcp` to retrieve VCP values from the
            DDC device instead of using this function directly.

            Args:
                amount: the number of bytes to read

            Raises:
                ValueError: if the read data is deemed invalid
            '''
            time.sleep(I2C.WAIT_TIME)

            ba = super().read(amount + 3)

            # check the bytes read
            checks = {
                'source address': ba[0] == I2C.DESTINATION_ADDR_W,
                'checksum': functools.reduce(operator.xor, ba) == I2C.HOST_ADDR_R,
                'length': len(ba) >= (ba[1] & ~self.PROTOCOL_FLAG) + 3
            }
            if False in checks.values():
                self.logger.error('i2c read check failed: ' + repr(checks))
                raise I2CValidationError(
                    'i2c read check failed: ' + repr(checks))

            return ba[2:-1]

        def getvcp(self, vcp_code: int) -> Tuple[int, int]:
            '''
            Retrieves a VCP value from the DDC device.

            Args:
                vcp_code: the VCP value to read, eg: `0x10` is brightness

            Returns:
                The current and maximum value respectively

            Raises:
                ValueError: if the read data is deemed invalid
            '''
            self.write(I2C.GET_VCP_CMD, vcp_code)
            ba = self.read(8)

            checks = {
                'is feature reply': ba[0] == I2C.GET_VCP_REPLY,
                'supported VCP opcode': ba[1] == 0,
                'answer matches request': ba[2] == vcp_code
            }
            if False in checks.values():
                self.logger.error('i2c read check failed: ' + repr(checks))
                raise I2CValidationError(
                    'i2c read check failed: ' + repr(checks))

            # current and max values
            return int.from_bytes(ba[6:8], 'big'), int.from_bytes(ba[4:6], 'big')

    @classmethod
    def get_display_info(cls, display: Optional[DisplayIdentifier] = None) -> List[dict]:
        all_displays = __cache__.get('i2c_display_info')
        if all_displays is None:
            all_displays = []
            index = 0

            for i2c_path in glob.glob('/dev/i2c-*'):
                if not os.path.exists(i2c_path):
                    continue

                try:
                    # open the I2C device using the host read address
                    device = cls.I2CDevice(i2c_path, cls.HOST_ADDR_R)
                    # read some 512 bytes from the device
                    data = device.read(512)
                except IOError as e:
                    cls.logger.error(
                        f'IOError reading from device {i2c_path}: {e}')
                    continue

                # search for the EDID header within our 512 read bytes
                start = data.find(bytes.fromhex('00 FF FF FF FF FF FF 00'))
                if start < 0:
                    continue

                # grab 128 bytes of the edid
                edid = data[start: start + 128]
                # parse the EDID
                manufacturer_id, manufacturer, model, name, serial = EDID.parse(
                    edid)
                # convert edid to hex string
                edid = ''.join(f'{i:02x}' for i in edid)

                all_displays.append(
                    {
                        'name': name,
                        'model': model,
                        'manufacturer': manufacturer,
                        'manufacturer_id': manufacturer_id,
                        'serial': serial,
                        'method': cls,
                        'index': index,
                        'edid': edid,
                        'i2c_bus': i2c_path
                    }
                )
                index += 1

            if all_displays:
                __cache__.store('i2c_display_info', all_displays, expires=2)

        if display is not None:
            return filter_monitors(display=display, haystack=all_displays, include=['i2c_bus'])
        return all_displays

    @classmethod
    def get_brightness(cls, display: Optional[int] = None) -> List[IntPercentage]:
        all_displays = cls.get_display_info()
        if display is not None:
            all_displays = [all_displays[display]]

        results = []
        for device in all_displays:
            interface = cls.DDCInterface(device['i2c_bus'])
            value, max_value = interface.getvcp(0x10)

            # make sure display's max brighness is cached
            cache_ident = '%s-%s-%s' % (device['name'],
                                        device['model'], device['serial'])
            if cache_ident not in cls._max_brightness_cache:
                cls._max_brightness_cache[cache_ident] = max_value
                cls.logger.info(
                    f'{cache_ident} max brightness:{max_value} (current: {value})')

            if max_value != 100:
                # if max value is not 100 then we have to adjust the scale to be
                # a percentage
                value = int((value / max_value) * 100)

            results.append(value)

        return results

    @classmethod
    def set_brightness(cls, value: IntPercentage, display: Optional[int] = None):
        all_displays = cls.get_display_info()
        if display is not None:
            all_displays = [all_displays[display]]

        for device in all_displays:
            # make sure display brightness max value is cached
            cache_ident = '%s-%s-%s' % (device['name'],
                                        device['model'], device['serial'])
            if cache_ident not in cls._max_brightness_cache:
                cls.get_brightness(display=device['index'])

            # scale the brightness value according to the max brightness
            max_value = cls._max_brightness_cache[cache_ident]
            if max_value != 100:
                value = int((value / 100) * max_value)

            interface = cls.DDCInterface(device['i2c_bus'])
            interface.setvcp(0x10, value)


class Light(BrightnessMethod):
    '''
    Wraps around [light](https://github.com/haikarainen/light), an external
    3rd party tool that can control brightness levels for e-DP displays.

    .. warning::
       As of April 2nd 2023, the official repository for the light project has
       [been archived](https://github.com/haikarainen/light/issues/147) and
       will no longer receive any updates unless another maintainer picks it
       up.
    '''

    executable: str = 'light'
    '''the light executable to be called'''

    @classmethod
    def get_display_info(cls, display: Optional[DisplayIdentifier] = None) -> List[dict]:
        '''
        Implements `BrightnessMethod.get_display_info`.

        Works by taking the output of `SysFiles.get_display_info` and
        filtering out any displays that aren't supported by Light
        '''
        light_output = check_output([cls.executable, '-L']).decode()
        displays = []
        index = 0
        for device in SysFiles.get_display_info():
            # SysFiles scrapes info from the same place that Light used to
            # so it makes sense to use that output
            if device['path'].replace('/sys/class', 'sysfs') in light_output:
                del device['scale']
                device['light_path'] = device['path'].replace(
                    '/sys/class', 'sysfs')
                device['method'] = cls
                device['index'] = index

                displays.append(device)
                index += 1

        if display is not None:
            displays = filter_monitors(display=display, haystack=displays, include=[
                                       'path', 'light_path'])
        return displays

    @classmethod
    def set_brightness(cls, value: IntPercentage, display: Optional[int] = None):
        info = cls.get_display_info()
        if display is not None:
            info = [info[display]]

        for i in info:
            check_output(
                f'{cls.executable} -S {value} -s {i["light_path"]}'.split(" "))

    @classmethod
    def get_brightness(cls, display: Optional[int] = None) -> List[IntPercentage]:
        info = cls.get_display_info()
        if display is not None:
            info = [info[display]]

        results = []
        for i in info:
            results.append(
                check_output([cls.executable, '-G', '-s', i['light_path']])
            )
        results = [int(round(float(i.decode()), 0)) for i in results]
        return results


class XRandr(BrightnessMethodAdv):
    '''collection of screen brightness related methods using the xrandr executable'''

    executable: str = 'xrandr'
    '''the xrandr executable to be called'''

    @classmethod
    def _gdi(cls):
        '''
        .. warning:: Don't use this
           This function isn't final and I will probably make breaking changes to it.
           You have been warned

        Gets all displays reported by XRandr even if they're not supported
        '''
        xrandr_output = check_output(
            [cls.executable, '--verbose']).decode().split('\n')

        display_count = 0
        tmp_display = {}

        for line_index, line in enumerate(xrandr_output):
            if line == '':
                continue

            if not line.startswith((' ', '\t')) and 'connected' in line and 'disconnected' not in line:
                if tmp_display:
                    yield tmp_display

                tmp_display = {
                    'name': line.split(' ')[0],
                    'interface': line.split(' ')[0],
                    'method': cls,
                    'index': display_count,
                    'model': None,
                    'serial': None,
                    'manufacturer': None,
                    'manufacturer_id': None,
                    'edid': None,
                    'unsupported': line.startswith('XWAYLAND')
                }
                display_count += 1

            elif 'EDID:' in line:
                # extract the edid from the chunk of the output that will contain the edid
                edid = ''.join(
                    i.replace('\t', '') for i in xrandr_output[line_index + 1: line_index + 9]
                )
                tmp_display['edid'] = edid

                for key, value in zip(
                    ('manufacturer_id', 'manufacturer', 'model', 'name', 'serial'),
                    EDID.parse(tmp_display['edid'])
                ):
                    if value is None:
                        continue
                    tmp_display[key] = value

            elif 'Brightness:' in line:
                tmp_display['brightness'] = int(
                    float(line.replace('Brightness:', '')) * 100)

        if tmp_display:
            yield tmp_display

    @classmethod
    def get_display_info(cls, display: Optional[DisplayIdentifier] = None, brightness: bool = False) -> List[dict]:
        '''
        Implements `BrightnessMethod.get_display_info`.

        Args:
            display: the index of the specific display to query.
                If unspecified, all detected displays are queried
            brightness: whether to include the current brightness
                in the returned info
        '''
        valid_displays = []
        for item in cls._gdi():
            if item['unsupported']:
                continue
            if not brightness:
                del item['brightness']
            del item['unsupported']
            valid_displays.append(item)
        if display is not None:
            valid_displays = filter_monitors(
                display=display, haystack=valid_displays, include=['interface'])
        return valid_displays

    @classmethod
    def get_brightness(cls, display: Optional[int] = None) -> List[IntPercentage]:
        monitors = cls.get_display_info(brightness=True)
        if display is not None:
            monitors = [monitors[display]]
        brightness = [i['brightness'] for i in monitors]

        return brightness

    @classmethod
    def set_brightness(cls, value: IntPercentage, display: Optional[int] = None):
        value = str(float(value) / 100)
        info = cls.get_display_info()
        if display is not None:
            info = [info[display]]

        for i in info:
            check_output([cls.executable, '--output',
                         i['interface'], '--brightness', value])


class DDCUtil(BrightnessMethodAdv):
    '''collection of screen brightness related methods using the ddcutil executable'''
    logger = logger.getChild('DDCUtil')

    executable: str = 'ddcutil'
    '''the ddcutil executable to be called'''
    sleep_multiplier: float = 0.5
    '''how long ddcutil should sleep between each DDC request (lower is shorter).
    See [the ddcutil docs](https://www.ddcutil.com/performance_options/) for more info.'''
    cmd_max_tries: int = 10
    '''max number of retries when calling the ddcutil'''
    enable_async = True
    _max_brightness_cache: dict = {}
    '''Cache for displays and their maximum brightness values'''

    @classmethod
    def _gdi(cls):
        '''
        .. warning:: Don't use this
           This function isn't final and I will probably make breaking changes to it.
           You have been warned

        Gets all displays reported by DDCUtil even if they're not supported
        '''
        raw_ddcutil_output = str(
            check_output(
                [
                    cls.executable, 'detect', '-v',
                    f'--sleep-multiplier={cls.sleep_multiplier}'
                ] + ['--async'] if cls.enable_async else [], max_tries=cls.cmd_max_tries
            )
        )[2:-1].split('\\n')
        # Use -v to get EDID string but this means output cannot be decoded.
        # Or maybe it can. I don't know the encoding though, so let's assume it cannot be decoded.
        # Use str()[2:-1] workaround

        # include "Invalid display" sections because they tell us where one displays metadata ends
        # and another begins. We filter out invalid displays later on
        ddcutil_output = [i for i in raw_ddcutil_output if i.startswith(
            ('Invalid display', 'Display', '\t', ' '))]
        tmp_display = {}
        display_count = 0

        for line_index, line in enumerate(ddcutil_output):
            if not line.startswith(('\t', ' ')):
                if tmp_display:
                    yield tmp_display

                tmp_display = {
                    'method': cls,
                    'index': display_count,
                    'model': None,
                    'serial': None,
                    'bin_serial': None,
                    'manufacturer': None,
                    'manufacturer_id': None,
                    'edid': None,
                    'unsupported': 'invalid display' in line.lower()
                }
                display_count += 1

            elif 'I2C bus' in line:
                tmp_display['i2c_bus'] = line[line.index('/'):]
                tmp_display['bus_number'] = int(
                    tmp_display['i2c_bus'].replace('/dev/i2c-', ''))

            elif 'Mfg id' in line:
                # Recently ddcutil has started reporting manufacturer IDs like
                # 'BNQ - UNK' or 'MSI - Microstep' so we have to split the line
                # into chunks of alpha chars and check for a valid mfg id
                for code in re.split(r'[^A-Za-z]', line.replace('Mfg id:', '').replace(' ', '')):
                    if len(code) != 3:
                        # all mfg ids are 3 chars long
                        continue

                    try:
                        (
                            tmp_display['manufacturer_id'],
                            tmp_display['manufacturer']
                        ) = _monitor_brand_lookup(code)
                    except TypeError:
                        continue
                    else:
                        break

            elif 'Model' in line:
                # the split() removes extra spaces
                name = line.replace('Model:', '').split()
                try:
                    tmp_display['model'] = name[1]
                except IndexError:
                    pass
                tmp_display['name'] = ' '.join(name)

            elif 'Serial number' in line:
                tmp_display['serial'] = line.replace(
                    'Serial number:', '').replace(' ', '') or None

            elif 'Binary serial number:' in line:
                tmp_display['bin_serial'] = line.split(' ')[-1][3:-1]

            elif 'EDID hex dump:' in line:
                try:
                    tmp_display['edid'] = ''.join(
                        ''.join(i.split()[1:17]) for i in ddcutil_output[line_index + 2: line_index + 10]
                    )
                except Exception:
                    pass

        if tmp_display:
            yield tmp_display

    @classmethod
    def get_display_info(cls, display: Optional[DisplayIdentifier] = None) -> List[dict]:
        valid_displays = __cache__.get('ddcutil_monitors_info')
        if valid_displays is None:
            valid_displays = []
            for item in cls._gdi():
                if item['unsupported']:
                    continue
                del item['unsupported']
                valid_displays.append(item)

            if valid_displays:
                __cache__.store('ddcutil_monitors_info', valid_displays)

        if display is not None:
            valid_displays = filter_monitors(
                display=display, haystack=valid_displays, include=['i2c_bus'])
        return valid_displays

    @classmethod
    def get_brightness(cls, display: Optional[int] = None) -> List[IntPercentage]:
        monitors = cls.get_display_info()
        if display is not None:
            monitors = [monitors[display]]

        res = []
        for monitor in monitors:
            value = __cache__.get(f'ddcutil_brightness_{monitor["index"]}')
            if value is None:
                cmd_out = check_output(
                    [
                        cls.executable,
                        'getvcp', '10', '-t',
                        '-b', str(monitor['bus_number']),
                        f'--sleep-multiplier={cls.sleep_multiplier}'
                    ], max_tries=cls.cmd_max_tries
                ).decode().split(' ')

                value = int(cmd_out[-2])
                max_value = int(cmd_out[-1])
                if max_value != 100:
                    # if the max brightness is not 100 then the number is not a percentage
                    # and will need to be scaled
                    value = int((value / max_value) * 100)

                # now make sure max brightness is recorded so set_brightness can use it
                cache_ident = '%s-%s-%s' % (monitor['name'],
                                            monitor['serial'], monitor['bin_serial'])
                if cache_ident not in cls._max_brightness_cache:
                    cls._max_brightness_cache[cache_ident] = max_value
                    cls.logger.debug(
                        f'{cache_ident} max brightness:{max_value} (current: {value})')

                __cache__.store(
                    f'ddcutil_brightness_{monitor["index"]}', value, expires=0.5)
            res.append(value)
        return res

    @classmethod
    def set_brightness(cls, value: IntPercentage, display: Optional[int] = None):
        monitors = cls.get_display_info()
        if display is not None:
            monitors = [monitors[display]]

        __cache__.expire(startswith='ddcutil_brightness_')
        for monitor in monitors:
            # check if monitor has a max brightness that requires us to scale this value
            cache_ident = '%s-%s-%s' % (monitor['name'],
                                        monitor['serial'], monitor['bin_serial'])
            if cache_ident not in cls._max_brightness_cache:
                cls.get_brightness(display=monitor['index'])

            if cls._max_brightness_cache[cache_ident] != 100:
                value = int((value / 100) * cls._max_brightness_cache[cache_ident])

            check_output(
                [
                    cls.executable, 'setvcp', '10', str(value),
                    '-b', str(monitor['bus_number']),
                    f'--sleep-multiplier={cls.sleep_multiplier}'
                ], max_tries=cls.cmd_max_tries
            )


def list_monitors_info(
    method: Optional[str] = None, allow_duplicates: bool = False, unsupported: bool = False
) -> List[dict]:
    '''
    Lists detailed information about all detected displays

    Args:
        method: the method the display can be addressed by. See `screen_brightness_control.get_methods`
            for more info on available methods
        allow_duplicates: whether to filter out duplicate displays (displays with the same EDID) or not
        unsupported: include detected displays that are invalid or unsupported
    '''
    all_methods = get_methods(method).values()
    haystack = []
    for method_class in all_methods:
        try:
            if unsupported and issubclass(method_class, BrightnessMethodAdv):
                haystack += method_class._gdi()
            else:
                haystack += method_class.get_display_info()
        except Exception as e:
            logger.warning(
                f'error grabbing display info from {method_class} - {format_exc(e)}')
            pass

    if allow_duplicates:
        return haystack

    try:
        # use filter_monitors to remove duplicates
        return filter_monitors(haystack=haystack)
    except NoValidDisplayError:
        return []



'''
# Without using a 3rd party program

As of version [0.13.0](https://github.com/Crozzers/screen_brightness_control/releases/tag/v0.13.0), you do not need a 3rd party program to adjust your display brightness.  
However, your user will need a few extra permissions.

## Laptop displays

Laptop displays are adjusted by reading and writing to the files in the `/sys/class/backlight` directory.
To avoid having to run this library as root every time, you can do the following:

1. Add a udev rule to allow editing of the brightness files.  
   This command adds a line to your backlight udev rules allowing you to write brightness values to the relevant files.

   ```bash
   echo 'SUBSYSTEM=="backlight",RUN+="/bin/chmod 666 /sys/class/backlight/%k/brightness /sys/class/backlight/%k/bl_power"' | sudo tee -a /etc/udev/rules.d/backlight-permissions.rules
   ```

2. Reload udev rules

   ```bash
   sudo udevadm control --reload-rules && udevadm trigger
   ```

3. If the above steps don't immediately work, reboot.

Credit to [Nayr438's comment](https://linustechtips.com/topic/1246132-allow-non-root-user-to-access-sysclassbacklight/?do=findComment&comment=14015728)
on the LTT forums for this.

## Desktop displays

Desktop displays are dealt with by reading and writing to the I2C buses, located at `/dev/i2c*` on a Linux system.
These are usually part of an `i2c` user group, which you can check by running the command `ls -lh /dev/i2c*`.  

If you add yourself to the `i2c` group your user should be able to read from and write
to the I2C buses without needing to use `sudo`. To add youself to the `i2c` group, run the following command:
```
usermod -a -G i2c [your username here]
```

The [I2C permissions section](https://www.ddcutil.com/i2c_permissions/) of the DDCUtil docs has some more details on this process.

# Using 3rd Party Programs

## Supported Programs

There are a number of external programs that `screen_brightness_control` can call upon if the native methods fail.
The advantage of using external programs is that, because they are installed using `sudo apt install ...`, they usually allow
users to adjust the backlight without having to manually fiddle with permissions. However, they do need to be installed.

Here is an outline of all of the external programs that `screen_brightness_control` can call upon:

Program       | Works on laptop displays | Works on external monitors | Per-display brightness control        | Requires Special Permissions After Install
--------------|--------------------------|----------------------------|---------------------------------------|-------------------------------------------------------------
ddcutil       | No                       | Yes (slowest) [1]          | Yes                                   | Read/write access for `/dev/i2c*` [2]
xrandr        | Yes                      | Yes           [3]          | Yes                                   | No
light         | Yes                      | No                         | Yes                                   | User must be in the `video` group [3]
[No program]  | Yes                      | Yes (slow)                 | Yes                                   | Read/write access for `/dev/i2c*` and `/sys/class/backlight`

#### Footnotes
[1] While both DDCUtil and the 1st party `linux.I2C` class do similar things over the same interface (I2C),
DDCUtil also supports communicating with monitors that implement the [Monitor Control Command Set over USB](https://www.ddcutil.com/usb)

[2] Read/write access for the `i2c` bus can be granted via the steps outlined in the [Desktop displays](#desktop-displays) section or by running the script using `sudo`.

[3] Xrandr does not actually change the backlight of the display, it just changes the brightness by applying a filter to the pixels to make them look dimmer/brighter.

[4] You can add yourself to the video user group by running `sudo usermod -a -G video [your username]` and then logging out and back in again for the changes to take effect.


## Install Instructions

* Arch
    * Xrandr: `sudo pacman -S xorg-xrandr`
    * DDCUtil: `sudo pacman -S ddcutil`
    * Light: `sudo pacman -S light-git`
* Debian/Ubuntu
    * XRandr: `sudo apt install x11-xserver-utils`
    * DDCUtil: `sudo apt install ddcutil`
    * Light: `sudo apt install light`
* Fedora
    * Xrandr: `sudo dnf install libXrandr`
    * DDCUtil: `sudo dnf install ddcutil`
    * Light: `sudo dnf install light`
'''



#!/bin/python
# flake8: noqa: E501

import os
import subprocess
import tarfile
from pathlib import Path
import shutil
import sys

from packaging import version

here = Path(__file__).parent
temp = here / 'temp'
os.makedirs(temp, exist_ok=True)

if '--skip-tag-pull' not in sys.argv:
    # get list of all versions from git tag
    subprocess.check_output(['git', 'pull', '--tags'])
all_versions = subprocess.check_output(['git', 'tag', '-l']).decode().rstrip('\n').split('\n')

# download any versions after v0.5.0
for v in filter(lambda a: version.Version(a) > version.Version('0.5.0'), all_versions):
    v = version.Version(v.replace('v', ''))
    if not os.path.isfile(temp / f'screen_brightness_control-{v}.tar.gz'):
        subprocess.check_output(['pip', 'download', '--no-deps', '--no-binary', ':all:', '-d', str(temp), f'screen_brightness_control=={v}'])

for file in os.listdir(temp):
    if os.path.isdir(temp / file):
        continue

    # extract each downloaded version
    with tarfile.open(temp / file) as t:
        out = temp / file.replace('screen_brightness_control-', '').replace('.tar.gz', '')
        if os.path.isdir(out):
            shutil.rmtree(out)

        def is_within_directory(directory, target):

            abs_directory = os.path.abspath(directory)
            abs_target = os.path.abspath(target)

            prefix = os.path.commonprefix([abs_directory, abs_target])

            return prefix == abs_directory

        def safe_extract(tar, path=".", members=None, *, numeric_owner=False):

            for member in tar.getmembers():
                member_path = os.path.join(path, member.name)
                if not is_within_directory(path, member_path):
                    raise Exception("Attempted Path Traversal in Tar File")

            tar.extractall(path, members, numeric_owner=numeric_owner) 


        safe_extract(t, out)

        # now find the __init__.py file in there and generate the docs for that module
        for root, _, files in os.walk(out):
            for f in files:
                if f == '__init__.py':
                    subprocess.check_output([sys.executable, str(here / 'make.py'), '-p', root])
                    break


'''
## get_brightness(`display=None, method=None, verbose_error=False`)
**Summary:**  
Gets the percentage brightness of all detected monitors. This is returned as a list of integers.  

**Arguments:**

* `display` - the specific display you wish to adjust. This can be an integer or a string (EDID, serial, name or model)
* `method` - the OS specific method to use. Use the [get_methods](#get_methods) function to get all available methods for your system.
* `verbose_error` - a boolean value to control how much detail any error messages should contain

**Usage:**  
```python
import screen_brightness_control as sbc

# get the current screen brightness (for all detected displays)
all_screens_brightness = sbc.get_brightness()
# get the brightness of the primary display
primary_display_brightness = sbc.get_brightness(display=0)[0]
# get the brightness of the secondary display (if connected)
secondary_display_brightness = sbc.get_brightness(display=1)[0]
# get the brightness for a named monitor
benq_brightness = sbc.get_brightness(display='BenQ GL2450H')[0]
```  


## set_brightness(`value, display=None, method=None, force=False, verbose_error=False, no_return=True`)
**Summary:**  
Sets the brightness to `value`. If `value` is a string and contains "+" or "-" then that value is added to/subtracted from the current brightness.  

**Arguments:**

* `value` - the level to set the brightness to. Can either be an integer or a string.
* `display` - the specific display you wish to adjust. This can be an integer or a string (EDID, serial, name or model)
* `method` - the OS specific method to use. Use the [get_methods](#get_methods) function to get all available methods for your system.
* `force` (Linux only) - if set to `False` then the brightness is never set to less than 1 because on Linux this often turns the screen off. If set to `True` then it will bypass this check
* `verbose_error` - a boolean value to control how much detail any error messages should contain
* `no_return` - whether this function should return the new brightness values. By default this behaviour is turned off

**Usage:**  
```python
import screen_brightness_control as sbc

#set brightness to 50%
sbc.set_brightness(50)

#set brightness to 0%
sbc.set_brightness(0, force=True)

#increase brightness by 25%
sbc.set_brightness('+25')

#decrease brightness by 30%
sbc.set_brightness('-30')

#set the brightness of display 0 to 50%
sbc.set_brightness(50, display=0)
```  


## fade_brightness(`finish, start=None, interval=0.01, increment=1, blocking=True, **kwargs`)
**Summary:**  
Fades the brightness from `start` to `finish` in steps of `increment`, pausing for `interval` seconds between each step.
If it runs in the main thread it will return the final brightness. Otherwise it returns the list of thread objects that the process is running in

**Arguments:**

* `finish` - The brightness value to fade to
* `start` - The value to start from. If not specified it defaults to the current brightness
* `interval` - The time interval between each step in brightness
* `increment` - The amount to change the brightness by each step in percent.
* `blocking` - If set to `False` it fades the brightness in a new thread
* `kwargs` - passed to `set_brightness`

**Usage:**  
```python
import screen_brightness_control as sbc

#fade brightness from the current brightness to 50%
sbc.fade_brightness(50)

#fade the brightness from 25% to 75%
sbc.fade_brightness(75, start=25)

#fade the brightness from the current value to 100% in steps of 10%
sbc.fade_brightness(100, increment=10)

#fade the brightness from 100% to 90% with time intervals of 0.1 seconds
sbc.fade_brightness(90, start=100, interval=0.1)

#fade the brightness to 100% in a new thread
sbc.fade_brightness(100, blocking=False)
```


## list_monitors(`method=None`)
**Summary:**  
Returns a list of the names of all detected monitors

**Arguments:**

* `method` - the OS specific method to use. Use the [get_methods](#get_methods) function to get all available methods for your system.

**Usage:**  
```python
import screen_brightness_control as sbc
monitor_names = sbc.list_monitors()
# eg: ['BenQ GL2450H', 'Dell U2211H']
```


## get_methods()
**Summary:**  
Returns a dictionary of brightness methods that you can use for adjusting screen brightness.

A method is just a class that uses a particular API or program to get display information and retrieve/set brightness levels for displays.
Each method may be able to address different kinds of displays (eg: laptop vs external monitors).

**Usage:**
```python
import screen_brightness_control as sbc

all_methods = sbc.get_methods()

for method_name, method_class in all_methods.items():
    print('Method:', method_name)
    print('Class:', method_class)
    print('Associated monitors:', sbc.list_monitors(method=method_name))
```
'''



'''
Script that generates fake importable modules to enable PDOC to import
modules that would otherwise not work in the current OS
'''
import importlib
import os
import shutil
import sys
from pathlib import Path

import pdoc

HERE = Path(__file__).parent
DUMMY_DIR = HERE / 'dummies'
DUMMY_TEMPLATE = """from dummy_imports import config_getattr
__getattr__ = config_getattr(__file__, __name__)
"""


def make_dummy_module(module: str):
    # this creates a dummy module using the template in './dummies/'
    os.makedirs(DUMMY_DIR, exist_ok=True)
    Path(DUMMY_DIR / '__init__.py').touch()

    # create directory
    os.makedirs(DUMMY_DIR / os.path.dirname(module), exist_ok=True)

    # create the file
    with open(DUMMY_DIR / module, 'w') as f:
        f.write(DUMMY_TEMPLATE)


def clear_dummy_modules():
    if not os.path.isdir(DUMMY_DIR):
        return
    shutil.rmtree(DUMMY_DIR)


def load_module(*args, **kwargs):
    # an override for pdoc.extract.load_module that tries
    # to fix ImportError and ModuleNotFoundError exceptions by creating
    # dummy modules to plug the gaps.
    last_err = None
    while True:
        try:
            return pdoc.extract._load_module(*args, **kwargs)
        except RuntimeError as e:
            err = e.__cause__
            if not isinstance(err, (ImportError, ModuleNotFoundError)):
                raise

            if last_err is not None and last_err == err.msg:
                # if we didn't manage to solve the error we had last time then
                # give up
                raise

            last_err = err.msg

            # get name of module it was trying to import
            if err.path is not None:
                module = err.path[err.path.index(err.name):]
            else:
                module = err.name.replace('.', '/')
                if not module.endswith('.py'):
                    module += '.py'
            # create dummy module
            make_dummy_module(module)
            # now load the created dummies into memory
            load_dummy_imports()


# override the default pdoc load_module function
pdoc.extract._load_module = pdoc.extract.load_module
pdoc.extract.load_module = load_module


def import_file(import_name, file):
    # import a module from a file path
    spec = importlib.util.spec_from_file_location(import_name, file)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def load_dummy_imports():
    # load all of our fake modules into sys.modules
    if str(DUMMY_DIR) not in sys.path:
        sys.path.insert(0, str(DUMMY_DIR))

    for root, _, files in os.walk(DUMMY_DIR):
        for file in files:
            if not file.endswith('.py'):
                continue

            if os.path.join(root, file) == __file__:
                continue

            if Path(os.path.join(root, file)) == DUMMY_DIR / '__init__.py':
                continue

            # get path to file to be imported
            file = Path(os.path.join(root, file))
            filepath = str(file.relative_to(DUMMY_DIR))
            # now get a module name, eg: ctypes/wintypes.py -> ctypes.wintypes
            sys_modules_name = filepath.replace('/__init__.py', '').replace('.py', '').replace('/', '.')

            # now import the module
            module = import_file(sys_modules_name, file)
            # and load it into the module list
            sys.modules[sys_modules_name] = module


class DummyObject():
    # a class that does nothing except look the part
    def __init__(self, __mod_name, __dummy_name, *args):
        self.__mod_name = __mod_name.replace('dummy_imports.', '')
        self.__dummy_name = __dummy_name

    def __repr__(self):
        # for representing objects in pdoc docs
        return f'{self.__mod_name}.{self.__dummy_name}'

    def __call__(*args):
        # compatibility for objects that might be callable
        pass

    def __mul__(self, item):
        # compatibility for stuff like `WCHAR * 128`
        return self


def config_getattr(__file__, __name__):
    # generates a __getattr__ function for dummy modules that figures out what
    # you want, makes something up and hands it back to you
    def func(name):
        modname = os.path.join(os.path.dirname(__file__), name + '.py')
        if os.path.isfile(modname):
            return import_file(name, modname)
        return DummyObject(__name__, name)
    return func



#!/bin/python
# flake8: noqa: E501
'''Script to generate documentation for this project.'''
import argparse
import glob
import json
import os
import shutil
import sys
from pathlib import Path

import pdoc
from jsmin import jsmin as minify_js
from packaging import version
from pdoc.render_helpers import minify_css

HERE = Path(__file__).parent
TEMPLATES = HERE / 'templates'
OUTPUT_DIR = HERE / 'docs'


def makedir(path: Path, destroy=False):
    if destroy:
        if os.path.isdir(path):
            shutil.rmtree(path)

    os.makedirs(path, exist_ok=True)


def get_directory_version(path: Path):
    if os.path.isfile(path / '_version.py'):
        with open(path / '_version.py', 'r') as f:
            contents = f.read()
    else:
        with open(path / '__init__.py', 'r') as f:
            contents = f.read()

    line = [i for i in contents.split('\n') if '__version__=' in i.replace(' ', '')][0]
    v = line.replace(' ', '').replace('__version__=', '').replace("'", "")

    return v


def configure_pdoc(**kwargs):
    if version.Version(pdoc.__version__) >= version.Version('12.2.0'):
        os.environ["PDOC_DEFINE_VIEW_SOURCE_MACRO"] = "1"
    else:
        os.environ["PDOC_DEFINE_VIEW_SOURCE_MACRO"] = "0"
    pdoc.render.configure(**{**PDOC_CONFIG, **kwargs})


def run_pdoc(source, output):
    if version.Version(pdoc.__version__) >= version.Version('12.2.0'):
        pdoc.pdoc(source, output_directory=output)
    else:
        pdoc.pdoc(source, output_directory=output, format='html')


def get_documentation_versions(directory):
    versions = [version.Version(i) for i in os.listdir(directory) if os.path.isdir(os.path.join(directory, i))]
    versions.sort(reverse=True)
    versions_grouped = {}

    while versions:
        item = versions.pop(0)
        group = [item]
        index = 0
        while index < len(versions):
            if versions[index].major == item.major and versions[index].minor == item.minor:
                group.append(versions.pop(index))
            else:
                index += 1
        group.sort()
        versions_grouped['docs/' + str(group[-1])] = [f'docs/{str(i)}' for i in reversed(group[:-1])]

    return versions_grouped


def custom_navigation_links():
    links = {}
    links['API Version'] = get_documentation_versions(OUTPUT_DIR / 'docs')

    for directory in os.listdir(HERE / 'source'):
        full_directory = HERE / 'source' / directory
        if not os.path.isdir(full_directory):
            continue

        category = directory.capitalize()
        links[category] = []
        for file in os.listdir(full_directory):
            if not os.path.isfile(full_directory / file):
                continue
            if file.startswith('__') or not file.endswith('.py'):
                continue
            links[category].append(f'{directory}/{file.rstrip(".py")}.html')
        if links[category]:
            links[category].sort()
        else:
            del links[category]

    return links


def latest_navigation_links():
    # get list of categories from `custom_navigation_links` that will need to be marked out of date.
    # basically any links containing a version number
    mark_latest = {}

    for category, links in custom_navigation_links().items():
        if 'version' in category.lower():
            mark_latest[category] = next(iter(links))

    return mark_latest


__version__ = get_directory_version(HERE / '..' / 'screen_brightness_control')

# pdoc.docstrings.GOOGLE_LIST_SECTIONS.extend(["Returns", "Yields"])
PDOC_CONFIG = dict(docformat='google', template_directory=TEMPLATES, footer_text=f'screen_brightness_control v{__version__}')
configure_pdoc()
pdoc.render.env.filters["minify_js"] = minify_js


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--path', help='Generate documentation for this path', default=str(HERE / '..' / 'screen_brightness_control'))
    parser.add_argument('--clean', help='Remove any existing documentation of any version', action='store_true')
    args = parser.parse_args()

    # use dummy imports for OS independent type hinting
    sys.path.insert(0, HERE)
    import dummy_imports
    dummy_imports.clear_dummy_modules()

    # generate top level documentation
    makedir(OUTPUT_DIR, destroy=args.clean)
    os.environ['BUILD_DOCS_LEVEL'] = 'homepage'
    configure_pdoc(search=False, show_source=False)
    run_pdoc(HERE / 'source', OUTPUT_DIR)

    # generate documentation for everything else in top level
    os.environ['BUILD_DOCS_LEVEL'] = 'extras'
    for dirname, files in custom_navigation_links().items():
        if dirname == 'API Version':
            continue

        makedir(OUTPUT_DIR / dirname.lower())
        for file in files:
            file = os.path.basename(file).replace('.html', '')
            module = pdoc.doc.Module.from_name(f'source.{dirname.lower()}.{file}')
            out = pdoc.render.html_module(module, [])
            with open(OUTPUT_DIR / dirname.lower() / f'{file}.html', 'w') as f:
                f.write(out)

    os.environ['BUILD_DOCS_LEVEL'] = 'api'
    configure_pdoc(search=True, show_source=True)
    os.makedirs(OUTPUT_DIR / 'docs', exist_ok=True)

    # generate documentation for specified path
    args.path = Path(args.path)
    path_version = get_directory_version(args.path)
    makedir(OUTPUT_DIR / 'docs' / path_version)
    configure_pdoc(footer_text=f'screen_brightness_control v{path_version}')
    run_pdoc(args.path, OUTPUT_DIR / 'docs' / path_version)

    # remove files for un-documented modules, like __main__ and _version
    for file in os.listdir(OUTPUT_DIR / 'docs' / path_version / 'screen_brightness_control'):
        if file.startswith('_') and file.endswith('.html'):
            os.remove(OUTPUT_DIR / 'docs' / path_version / 'screen_brightness_control' / file)

    # clear up dummy imports
    dummy_imports.clear_dummy_modules()

    # read version switcher js
    with open(TEMPLATES / 'version_navigator.js', 'r') as f:
        js_code = f.read()

    # insert list of navigation links
    navigation_links = json.dumps(custom_navigation_links())
    dated_links = json.dumps(latest_navigation_links())
    js_code = js_code.replace('var all_nav_links = {};', f'var all_nav_links = {navigation_links};')
    js_code = js_code.replace('var mark_latest = {};', f'var mark_latest = {dated_links};')

    # write to gh-pages dir
    with open(OUTPUT_DIR / 'version_navigator.js', 'w') as f:
        f.write(minify_js(js_code))

    # copy over css file
    with open(TEMPLATES / 'version_navigator.css', 'r') as f:
        # write to gh-pages dir
        with open(OUTPUT_DIR / 'version_navigator.css', 'w') as g:
            g.write(minify_css(f.read()))

    # remove top-level search.js as it is not used
    if os.path.isfile(OUTPUT_DIR / 'search.js'):
        os.remove(OUTPUT_DIR / 'search.js')



'''
These are some example projects you can create using this library. I recommend reading
the [Quick Start Guide](Quick%20Start%20Guide.html) for more details on the main functions
used in these examples.

## Minimal UI
Description: A basic tkinter user interface for changing monitor brightness  
Author: [Michel Weinachter](https://fr.linkedin.com/in/michel-weinachter-a212931)  
License: [MIT](https://mit-license.org/)  

```python
import tkinter as tk
import screen_brightness_control as sbc

def set_values(j):
    print(j)
    for i in monitors_list :
        a, b = i
        sbc.set_brightness(value=b.get(),display=a)

def updateValue(self, event):
        print(self.slider.get())

raw_monitors_list= sbc.list_monitors()
monitors_list = []

master = tk.Tk()
master.title("Screen brightness control")

for i in raw_monitors_list :
    print(i)
    sbc.get_brightness(i)
    w = tk.Label(master, text=i)
    w.pack()
    x = tk.Scale(master, from_=0, to=100, orient="horizontal", length=350,resolution=1, tickinterval=10, command=set_values)
    x.set(sbc.get_brightness(i))
    x.pack()
    monitors_list.append((i, x))


tk.mainloop()
```


## Auto Night Light
Description: Automatically dim your display brightness when it gets late in the evening  
Author: [Crozzers](https://github.com/Crozzers)  
License: [MIT](https://mit-license.org/)  
```python
import time
import screen_brightness_control as sbc
from datetime import datetime

while True:
    now = datetime.now()
    if now.hour >= 19:
        # if it is after 7pm, turn down the brightness to 50%
        sbc.fade_brightness(50)
    elif now.hour >= 21:
        # if it is after 9pm, turn down the brightness to 25%
        sbc.fade_brightness(25)
    elif now.hour >= 9:
        # if it is after 9am, turn up the brightness to 100%
        sbc.fade_brightness(100)

    time.sleep(60)
```
'''


import logging
import platform
import threading
import time
import traceback
import warnings
from dataclasses import dataclass, fields, field
from typing import Any, Dict, List, Optional, Tuple, Union

from ._debug import info as debug_info  # noqa: F401
from ._version import __author__, __version__  # noqa: F401
from .exceptions import NoValidDisplayError, format_exc  # noqa: F401
from .helpers import (MONITOR_MANUFACTURER_CODES,  # noqa: F401
                      BrightnessMethod, ScreenBrightnessError,
                      logarithmic_range, percentage)
from .types import DisplayIdentifier, IntPercentage, Percentage

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def get_brightness(
    display: Optional[DisplayIdentifier] = None,
    method: Optional[str] = None,
    verbose_error: bool = False
) -> List[IntPercentage]:
    '''
    Returns the current brightness of one or more displays

    Args:
        display (types.DisplayIdentifier): the specific display to query
        method: the method to use to get the brightness. See `get_methods` for
            more info on available methods
        verbose_error: controls the level of detail in the error messages

    Returns:
        A list of `IntPercentage` values, each being the brightness of an
        individual display. Invalid displays may return None.

    Example:
        ```python
        import screen_brightness_control as sbc

        # get the current screen brightness (for all detected displays)
        current_brightness = sbc.get_brightness()

        # get the brightness of the primary display
        primary_brightness = sbc.get_brightness(display=0)

        # get the brightness of the secondary display (if connected)
        secondary_brightness = sbc.get_brightness(display=1)
        ```
    '''
    return __brightness(display=display, method=method, meta_method='get', verbose_error=verbose_error)


def set_brightness(
    value: Percentage,
    display: Optional[DisplayIdentifier] = None,
    method: Optional[str] = None,
    force: bool = False,
    verbose_error: bool = False,
    no_return: bool = True
) -> Optional[List[IntPercentage]]:
    '''
    Sets the brightness level of one or more displays to a given value.

    Args:
        value (types.Percentage): the new brightness level
        display (types.DisplayIdentifier): the specific display to adjust
        method: the method to use to set the brightness. See `get_methods` for
            more info on available methods
        force: [*Linux Only*] if False the brightness will never be set lower than 1.
            This is because on most displays a brightness of 0 will turn off the backlight.
            If True, this check is bypassed
        verbose_error: boolean value controls the amount of detail error messages will contain
        no_return: don't return the new brightness level(s)

    Returns:
        If `no_return` is set to `True` (the default) then this function returns nothing.
        Otherwise, a list of `types.IntPercentage` is returned, each item being the new
        brightness of each adjusted display (invalid displays may return None)

    Example:
        ```python
        import screen_brightness_control as sbc

        # set brightness to 50%
        sbc.set_brightness(50)

        # set brightness to 0%
        sbc.set_brightness(0, force=True)

        # increase brightness by 25%
        sbc.set_brightness('+25')

        # decrease brightness by 30%
        sbc.set_brightness('-30')

        # set the brightness of display 0 to 50%
        sbc.set_brightness(50, display=0)
        ```
    '''
    if isinstance(value, str) and ('+' in value or '-' in value):
        output = []
        for monitor in filter_monitors(display=display, method=method):
            identifier = Monitor.get_identifier(monitor)[1]
            current_value = get_brightness(display=identifier)[0]
            result = set_brightness(
                # don't need to calculate lower bound here because it will be
                # done by the other path in `set_brightness`
                percentage(value, current=current_value),
                display=identifier,
                force=force,
                verbose_error=verbose_error,
                no_return=no_return
            )
            if result is None:
                output.append(result)
            else:
                output += result

        return output

    if platform.system() == 'Linux' and not force:
        lower_bound = 1
    else:
        lower_bound = 0

    value = percentage(value, lower_bound=lower_bound)

    return __brightness(
        value, display=display, method=method,
        meta_method='set', no_return=no_return,
        verbose_error=verbose_error
    )


def fade_brightness(
    finish: Percentage,
    start: Optional[Percentage] = None,
    interval: float = 0.01,
    increment: int = 1,
    blocking: bool = True,
    force: bool = False,
    logarithmic: bool = True,
    **kwargs
) -> Union[List[threading.Thread], List[IntPercentage]]:
    '''
    Gradually change the brightness of one or more displays

    Args:
        finish (types.Percentage): fade to this brightness level
        start (types.Percentage): where the brightness should fade from.
            If this arg is not specified, the fade will be started from the
            current brightness.
        interval: the time delay between each step in brightness
        increment: the amount to change the brightness by per step
        blocking: whether this should occur in the main thread (`True`) or a new daemonic thread (`False`)
        force: [*Linux Only*] if False the brightness will never be set lower than 1.
            This is because on most displays a brightness of 0 will turn off the backlight.
            If True, this check is bypassed
        logarithmic: follow a logarithmic brightness curve when adjusting the brightness
        **kwargs: passed through to `filter_monitors` for display selection.
            Will also be passed to `get_brightness` if `blocking is True`

    Returns:
        By default, this function calls `get_brightness()` to return the new
        brightness of any adjusted displays.

        If the `blocking` is set to `False`, then a list of threads are
        returned, one for each display being faded.

    Example:
        ```python
        import screen_brightness_control as sbc

        # fade brightness from the current brightness to 50%
        sbc.fade_brightness(50)

        # fade the brightness from 25% to 75%
        sbc.fade_brightness(75, start=25)

        # fade the brightness from the current value to 100% in steps of 10%
        sbc.fade_brightness(100, increment=10)

        # fade the brightness from 100% to 90% with time intervals of 0.1 seconds
        sbc.fade_brightness(90, start=100, interval=0.1)

        # fade the brightness to 100% in a new thread
        sbc.fade_brightness(100, blocking=False)
        ```
    '''
    # make sure only compatible kwargs are passed to filter_monitors
    available_monitors = filter_monitors(
        **{k: v for k, v in kwargs.items() if k in (
            'display', 'haystack', 'method', 'include'
        )}
    )

    threads = []
    for i in available_monitors:
        monitor = Monitor(i)

        t1 = monitor.fade_brightness(finish, start=start, interval=interval, increment=increment,
                                     force=force, logarithmic=logarithmic, blocking=False)
        threads.append(t1)

    if not blocking:
        return threads

    for t in threads:
        t.join()
    return get_brightness(**kwargs)


def list_monitors_info(
    method: Optional[str] = None, allow_duplicates: bool = False, unsupported: bool = False
) -> List[dict]:
    '''
    List detailed information about all displays that are controllable by this library

    Args:
        method: the method to use to list the available displays. See `get_methods` for
            more info on available methods
        allow_duplicates: whether to filter out duplicate displays or not
        unsupported: include detected displays that are invalid or unsupported

    Returns:
        list: list of dictionaries containing information about the detected displays

    Example:
        ```python
        import screen_brightness_control as sbc
        displays = sbc.list_monitors_info()
        for display in displays:
            print('=======================')
            # the manufacturer name plus the model
            print('Name:', display['name'])
            # the general model of the display
            print('Model:', display['model'])
            # the serial of the display
            print('Serial:', display['serial'])
            # the name of the brand of the display
            print('Manufacturer:', display['manufacturer'])
            # the 3 letter code corresponding to the brand name, EG: BNQ -> BenQ
            print('Manufacturer ID:', display['manufacturer_id'])
            # the index of that display FOR THE SPECIFIC METHOD THE DISPLAY USES
            print('Index:', display['index'])
            # the method this display can be addressed by
            print('Method:', display['method'])
            # the EDID string associated with that display
            print('EDID:', display['edid'])
        ```
    '''
    return _OS_MODULE.list_monitors_info(
        method=method, allow_duplicates=allow_duplicates, unsupported=unsupported
    )


def list_monitors(method: Optional[str] = None) -> List[str]:
    '''
    List the names of all detected displays

    Args:
        method: the method to use to list the available displays. See `get_methods` for
            more info on available methods

    Example:
        ```python
        import screen_brightness_control as sbc
        display_names = sbc.list_monitors()
        # eg: ['BenQ GL2450H', 'Dell U2211H']
        ```
    '''
    return [i['name'] for i in list_monitors_info(method=method)]


def get_methods(name: str = None) -> Dict[str, BrightnessMethod]:
    '''
    Returns all available brightness method names and their associated classes.

    Args:
        name: if specified, return the method corresponding to this name

    Raises:
        ValueError: if the given name is incorrect

    Example:
        ```python
        import screen_brightness_control as sbc

        all_methods = sbc.get_methods()

        for method_name, method_class in all_methods.items():
            print('Method:', method_name)
            print('Class:', method_class)
            print('Associated monitors:', sbc.list_monitors(method=method_name))
        ```
    '''
    methods = {i.__name__.lower(): i for i in _OS_METHODS}

    if name is None:
        return methods

    if not isinstance(name, str):
        raise TypeError(f'name must be of type str, not {type(name)!r}')

    name = name.lower()
    if name in methods:
        return {name: methods[name]}

    logger.debug(f'requested method {name!r} invalid')
    raise ValueError(
        f'invalid method {name!r}, must be one of: {list(methods)}')


@dataclass
class Display():
    '''
    Represents a single connected display.
    '''
    index: int
    '''The index of the display relative to the method it uses.
    So if the index is 0 and the method is `windows.VCP`, then this is the 1st
    display reported by `windows.VCP`, not the first display overall.'''
    method: BrightnessMethod
    '''The method by which this monitor can be addressed.
    This will be a class from either the windows or linux sub-module'''

    edid: str = None
    '''A 256 character hex string containing information about a display and its capabilities'''
    manufacturer: str = None
    '''Name of the display's manufacturer'''
    manufacturer_id: str = None
    '''3 letter code corresponding to the manufacturer name'''
    model: str = None
    '''Model name of the display'''
    name: str = None
    '''The name of the display, often the manufacturer name plus the model name'''
    serial: str = None
    '''The serial number of the display or (if serial is not available) an ID assigned by the OS'''

    _logger: logging.Logger = field(init=False)

    def __post_init__(self):
        self._logger = logger.getChild(self.__class__.__name__).getChild(
            str(self.get_identifier()[1])[:20])

    def fade_brightness(
        self,
        finish: Percentage,
        start: Optional[Percentage] = None,
        interval: float = 0.01,
        increment: int = 1,
        force: bool = False,
        logarithmic: bool = True
    ) -> IntPercentage:
        '''
        Gradually change the brightness of this display to a set value.
        This works by incrementally changing the brightness until the desired
        value is reached.

        Args:
            finish (types.Percentage): the brightness level to end up on
            start (types.Percentage): where the fade should start from. Defaults
                to whatever the current brightness level for the display is
            interval: time delay between each change in brightness
            increment: amount to change the brightness by each time (as a percentage)
            force: [*Linux only*] allow the brightness to be set to 0. By default,
                brightness values will never be set lower than 1, since setting them to 0
                often turns off the backlight
            logarithmic: follow a logarithmic curve when setting brightness values.
                See `logarithmic_range` for rationale

        Returns:
            The brightness of the display after the fade is complete.
            See `types.IntPercentage`
        '''
        # minimum brightness value
        if platform.system() == 'Linux' and not force:
            lower_bound = 1
        else:
            lower_bound = 0

        current = self.get_brightness()

        finish = percentage(finish, current, lower_bound)
        start = percentage(
            current if start is None else start, current, lower_bound)

        range_func = logarithmic_range if logarithmic else range
        increment = abs(increment)
        if start > finish:
            increment = -increment

        self._logger.debug(
            f'fade {start}->{finish}:{increment}:logarithmic={logarithmic}')

        for value in range_func(start, finish, increment):
            self.set_brightness(value)
            time.sleep(interval)

        if self.get_brightness() != finish:
            self.set_brightness(finish, no_return=True)

        return self.get_brightness()

    @classmethod
    def from_dict(cls, display: dict) -> 'Display':
        '''
        Initialise an instance of the class from a dictionary, ignoring
        any unwanted keys
        '''
        return cls(
            index=display['index'],
            method=display['method'],
            edid=display['edid'],
            manufacturer=display['manufacturer'],
            manufacturer_id=display['manufacturer_id'],
            model=display['model'],
            name=display['name'],
            serial=display['serial']
        )

    def get_brightness(self) -> IntPercentage:
        '''
        Returns the brightness of this display.

        Returns:
            The brightness value of the display, as a percentage.
            See `types.IntPercentage`
        '''
        return self.method.get_brightness(display=self.index)[0]

    def get_identifier(self) -> Tuple[str, DisplayIdentifier]:
        '''
        Returns the `types.DisplayIdentifier` for this display.
        Will iterate through the EDID, serial, name and index and return the first
        value that is not equal to None

        Returns:
            The name of the property returned and the value of said property.
            EG: `('serial', '123abc...')` or `('name', 'BenQ GL2450H')`
        '''
        for key in ('edid', 'serial', 'name', 'index'):
            value = getattr(self, key, None)
            if value is not None:
                return key, value

    def is_active(self) -> bool:
        '''
        Attempts to retrieve the brightness for this display. If it works the display is deemed active
        '''
        try:
            self.get_brightness()
            return True
        except Exception as e:
            self._logger.error(
                f'Monitor.is_active: {self.get_identifier()} failed get_brightness call'
                f' - {format_exc(e)}'
            )
            return False

    def set_brightness(self, value: Percentage, force: bool = False):
        '''
        Sets the brightness for this display. See `set_brightness` for the full docs

        Args:
            value (types.Percentage): the brightness percentage to set the display to
            force: allow the brightness to be set to 0 on Linux. This is disabled by default
                because setting the brightness of 0 will often turn off the backlight
        '''
        # convert brightness value to percentage
        if platform.system() == 'Linux' and not force:
            lower_bound = 1
        else:
            lower_bound = 0

        value = percentage(
            value,
            current=lambda: self.method.get_brightness(display=self.index)[0],
            lower_bound=lower_bound
        )

        self.method.set_brightness(value, display=self.index)


class Monitor(Display):
    '''
    Legacy class for managing displays.

    .. warning:: Deprecated
       Deprecated for removal in v0.23.0. Please use the new `Display` class instead
    '''

    def __init__(self, display: Union[int, str, dict]):
        '''
        Args:
            display (types.DisplayIdentifier or dict): the display you
                wish to control. Is passed to `filter_monitors`
                to decide which display to use.

        Example:
            ```python
            import screen_brightness_control as sbc

            # create a class for the primary display and then a specifically named monitor
            primary = sbc.Monitor(0)
            benq_monitor = sbc.Monitor('BenQ GL2450H')

            # check if the benq monitor is the primary one
            if primary.serial == benq_monitor.serial:
                print('BenQ GL2450H is the primary display')
            else:
                print('The primary display is', primary.name)

            # DEPRECATED BEHAVIOUR
            # Will be removed in v0.22.0
            print(primary['name'])
            print(benq_monitor['name'])
            ```
        '''
        warnings.warn(
            (
                '`Monitor` is deprecated for removal in v0.23.0.'
                ' Please use the new `Display` class instead'
            ),
            DeprecationWarning
        )

        monitors_info = list_monitors_info(allow_duplicates=True)
        if isinstance(display, dict):
            if display in monitors_info:
                info = display
            else:
                info = filter_monitors(
                    display=self.get_identifier(display),
                    haystack=monitors_info
                )[0]
        else:
            info = filter_monitors(display=display, haystack=monitors_info)[0]

        # make a copy so that we don't alter the dict in-place
        info = info.copy()

        kw = [i.name for i in fields(Display) if i.init]
        super().__init__(**{k: v for k, v in info.items() if k in kw})

        # this assigns any extra info that is returned to this class
        # eg: the 'interface' key in XRandr monitors on Linux
        for key, value in info.items():
            if key not in kw and value is not None:
                setattr(self, key, value)

    def __getitem__(self, item: Any) -> Any:
        '''
        .. warning:: Deprecated
           This behaviour is deprecated and will be removed in v0.22.0
        '''
        warnings.warn(
            '`Monitor.__getitem__` is deprecated for removal in v0.22.0',
            DeprecationWarning
        )
        return getattr(self, item)

    def get_identifier(self, monitor: dict = None) -> Tuple[str, DisplayIdentifier]:
        '''
        Returns the `types.DisplayIdentifier` for this display.
        Will iterate through the EDID, serial, name and index and return the first
        value that is not equal to None

        Args:
            monitor: extract an identifier from this dict instead of the monitor class

        Returns:
            A tuple containing the name of the property returned and the value of said
            property. EG: `('serial', '123abc...')` or `('name', 'BenQ GL2450H')`

        Example:
            ```python
            import screen_brightness_control as sbc
            primary = sbc.Monitor(0)
            print(primary.get_identifier())  # eg: ('serial', '123abc...')

            secondary = sbc.list_monitors_info()[1]
            print(primary.get_identifier(monitor=secondary))  # eg: ('serial', '456def...')

            # you can also use the class uninitialized
            print(sbc.Monitor.get_identifier(secondary))  # eg: ('serial', '456def...')
            ```
        '''
        if monitor is None:
            if isinstance(self, dict):
                monitor = self
            else:
                return super().get_identifier()

        for key in ('edid', 'serial', 'name', 'index'):
            value = monitor[key]
            if value is not None:
                return key, value

    def set_brightness(
        self,
        value: Percentage,
        no_return: bool = True,
        force: bool = False
    ) -> Optional[IntPercentage]:
        '''
        Wrapper for `Display.set_brightness`

        Args:
            value: see `Display.set_brightness`
            no_return: do not return the new brightness after it has been set
            force: see `Display.set_brightness`
        '''
        # refresh display info, in case another display has been unplugged or something
        # which would change the index of this display
        self.get_info()
        super().set_brightness(value, force)
        if not no_return:
            return self.get_brightness()

    def get_brightness(self) -> IntPercentage:
        # refresh display info, in case another display has been unplugged or something
        # which would change the index of this display
        self.get_info()
        return super().get_brightness()

    def fade_brightness(
        self,
        *args,
        blocking: bool = True,
        **kwargs
    ) -> Union[threading.Thread, IntPercentage]:
        '''
        Wrapper for `Display.fade_brightness`

        Args:
            *args: see `Display.fade_brightness`
            blocking: run this function in the current thread and block until
                it completes. If `False`, the fade will be run in a new daemonic
                thread, which will be started and returned
            **kwargs: see `Display.fade_brightness`
        '''
        if not blocking:
            result = threading.Thread(
                target=super().fade_brightness, args=args, kwargs=kwargs, daemon=True)
            result.start()
        else:
            result = super().fade_brightness(*args, **kwargs)

        return result

    @classmethod
    def from_dict(cls, display) -> 'Monitor':
        return cls(display)

    def get_info(self, refresh: bool = True) -> dict:
        '''
        Returns all known information about this monitor instance

        Args:
            refresh: whether to refresh the information
                or to return the cached version

        Example:
            ```python
            import screen_brightness_control as sbc

            # initialize class for primary display
            primary = sbc.Monitor(0)
            # get the info
            info = primary.get_info()
            ```
        '''
        def vars_self():
            return {k: v for k, v in vars(self).items() if not k.startswith('_')}

        if not refresh:
            return vars_self()

        identifier = self.get_identifier()

        if identifier is not None:
            # refresh the info we have on this monitor
            info = filter_monitors(
                display=identifier[1], method=self.method.__name__)[0]
            for key, value in info.items():
                if value is not None:
                    setattr(self, key, value)

        return vars_self()


def filter_monitors(
    display: Optional[DisplayIdentifier] = None,
    haystack: Optional[List[dict]] = None,
    method: Optional[str] = None,
    include: List[str] = []
) -> List[dict]:
    '''
    Searches through the information for all detected displays
    and attempts to return the info matching the value given.
    Will attempt to match against index, name, model, edid, method and serial

    Args:
        display (types.DisplayIdentifier): the display you are searching for
        haystack: the information to filter from.
            If this isn't set it defaults to the return of `list_monitors_info`
        method: the method the monitors use. See `get_methods` for
            more info on available methods
        include: extra fields of information to sort by

    Raises:
        NoValidDisplayError: if the display does not have a match

    Example:
        ```python
        import screen_brightness_control as sbc

        search = 'GL2450H'
        match = sbc.filter_displays(search)
        print(match)
        # EG output: [{'name': 'BenQ GL2450H', 'model': 'GL2450H', ... }]
        ```
    '''
    if display is not None and type(display) not in (str, int):
        raise TypeError(
            f'display kwarg must be int or str, not "{type(display).__name__}"')

    def get_monitor_list():
        # if we have been provided with a list of monitors to sift through then use that
        # otherwise, get the info ourselves
        if haystack is not None:
            monitors_with_duplicates = haystack
            if method is not None:
                method_class = next(get_methods(method).values())
                monitors_with_duplicates = [
                    i for i in haystack if i['method'] == method_class]
        else:
            monitors_with_duplicates = list_monitors_info(
                method=method, allow_duplicates=True)

        return monitors_with_duplicates

    def filter_monitor_list(to_filter):
        # This loop does two things:
        # 1. Filters out duplicate monitors
        # 2. Matches the display kwarg (if applicable)
        filtered_displays = {}
        for monitor in to_filter:
            # find a valid identifier for a monitor, excluding any which are equal to None
            added = False
            for identifier in ['edid', 'serial', 'name', 'model'] + include:
                # check we haven't already added the monitor
                if monitor.get(identifier, None) is None:
                    continue

                m_id = monitor[identifier]
                if m_id in filtered_displays:
                    break

                if isinstance(display, str) and m_id != display:
                    continue

                if not added:
                    filtered_displays[m_id] = monitor
                    added = True

                # if the display kwarg is an integer and we are currently at that index
                if isinstance(display, int) and len(filtered_displays) - 1 == display:
                    return [monitor]

                if added:
                    break
        return list(filtered_displays.values())

    duplicates = []
    for _ in range(3):
        duplicates = get_monitor_list()
        if duplicates:
            break
        time.sleep(0.4)
    else:
        msg = 'no displays detected'
        if method is not None:
            msg += f' with method: {method!r}'
        raise NoValidDisplayError(msg)

    monitors = filter_monitor_list(duplicates)
    if not monitors:
        # if no displays matched the query
        msg = 'no displays found'
        if display is not None:
            msg += f' with name/serial/model/edid/index of {display!r}'
        if method is not None:
            msg += f' with method of {method!r}'
        raise NoValidDisplayError(msg)

    return monitors


def __brightness(
    *args, display=None, method=None, meta_method='get', no_return=False,
    verbose_error=False, **kwargs
):
    '''Internal function used to get/set brightness'''
    logger.debug(
        f"brightness {meta_method} request display {display} with method {method}")

    output = []
    errors = []
    method = method.lower() if isinstance(method, str) else method

    for monitor in filter_monitors(display=display, method=method):
        try:
            if meta_method == 'set':
                monitor['method'].set_brightness(
                    *args, display=monitor['index'], **kwargs)
                if no_return:
                    output.append(None)
                    continue

            output += monitor['method'].get_brightness(
                display=monitor['index'], **kwargs)
        except Exception as e:
            output.append(None)
            errors.append((
                monitor, e.__class__.__name__,
                traceback.format_exc() if verbose_error else e
            ))

    if output:
        output_is_none = set(output) == {None}
        if (
            # can't have None output if we are trying to get the brightness
            (meta_method == 'get' and not output_is_none)
            or (
                # if we are setting the brightness then we CAN have a None output
                # but only if no_return is True.
                meta_method == 'set'
                and ((no_return and output_is_none) or not output_is_none)
            )
        ):
            return None if no_return else output

    # if the function hasn't returned then it has failed
    msg = '\n'
    if errors:
        for monitor, exc_name, exc in errors:
            if isinstance(monitor, str):
                msg += f'\t{monitor}'
            else:
                msg += f'\t{monitor["name"]} ({monitor["serial"]})'
            msg += f' -> {exc_name}: '
            msg += str(exc).replace('\n', '\n\t\t') + '\n'
    else:
        msg += '\tno valid output was received from brightness methods'

    raise ScreenBrightnessError(msg)


if platform.system() == 'Windows':
    from . import windows
    _OS_MODULE = windows
    _OS_METHODS = (_OS_MODULE.WMI, _OS_MODULE.VCP)
elif platform.system() == 'Linux':
    from . import linux
    _OS_MODULE = linux
    _OS_METHODS = (
        _OS_MODULE.SysFiles, _OS_MODULE.I2C,
        _OS_MODULE.XRandr, _OS_MODULE.DDCUtil,
        _OS_MODULE.Light
    )
else:
    logger.warning(
        f'package imported on unsupported platform ({platform.system()})')



import os
import platform
import sys


def run_test(file, synthetic=False):
    print(f'Run test: {file}, synthetic={synthetic}')
    exitcode = os.system(
        '"%s" %s %s' % (
            sys.executable, file,
            '--synthetic' if synthetic else ''
        )
    )
    if exitcode:
        raise Exception(f'Test failed with exit code {exitcode}')


if __name__ == '__main__':
    if '--synthetic' in sys.argv:
        run_test('tests/test_init.py', synthetic=True)
        run_test('tests/test_helpers.py', synthetic=True)
    else:
        for file in sorted(os.listdir('tests')):
            if not file.startswith('test_'):
                continue
            if not file.endswith('.py'):
                continue

            file = os.path.join('tests/', file)

            if any(i in file for i in ('windows', 'linux', 'darwin')):
                if platform.system().lower() not in file:
                    print(f'Skip: {file} due to OS requirements')
                    continue

            run_test(file)



import subprocess


def format_exc(e: Exception) -> str:
    return f'{type(e).__name__}: {e}'


class ScreenBrightnessError(Exception):
    '''
    Generic error class designed to make catching errors under one umbrella easy.
    '''
    ...


class EDIDParseError(ScreenBrightnessError):
    ...


class NoValidDisplayError(ScreenBrightnessError, LookupError):
    ...


class I2CValidationError(ScreenBrightnessError):
    ...


class MaxRetriesExceededError(ScreenBrightnessError, subprocess.CalledProcessError):
    ...



'''
## General FAQ
### Why doesn't the library support Mac?
I don't have a Mac to develop the library on so I have no way to test if the code works.
If you have a Mac and you would like to contribute, please do not hesitate to [create a pull request](https://github.com/Crozzers/screen_brightness_control/pulls).

### Library throws exception - `NameError: name '_OS_MODULE' is not defined`
This error is thown when attempting to use the library on unsupported platforms.

Upon import, the library will check which OS it is running on and will attempt to import the
relevant, OS specific sub-modules. When running on an unsupported OS, this will not be possible and
so `_OS_MODULE` will not be defined.

In the past, a `NotImplementedError` would have been raised at this stage, letting you know the issue.
This behaviour was changed in v0.16.0 to simply log a warning.

### My monitor isn't supported

This library supports most laptop displays and desktop monitors. Desktop monitors must support DDC/CI and have it
enabled (try checking your monitor's on screen display menu).

If your display meets these conditions but isn't showing up, you can check if the library doesn't support your
display like so:

```python
import screen_brightness_control as sbc
# print all detected but unsupported monitors
for display in sbc.list_monitors_info(allow_duplicates=True, unsupported=True):
    if display.get('unsupported'):
        print(display)
```

## Windows FAQ
### When I call `get_brightness()` the returned value isn't what I set it to
Not all monitors can set the brightness for every value between 0 and 100. Some of them have a number of 'levels' that they can be set to.
You can likely see this if you open your display settings and very slowly move the brightness slider.  
You can find out your brightness 'levels' by running the following python code:
```python
import wmi
monitor = wmi.WMI(namespace='wmi').WmiMonitorBrightness()[0]
#the number of levels the monitor can be set to
print(monitor.Levels)
#the actual brightness values your monitor can be set to
print(monitor.Level)
```


## Linux FAQ
### Why do I always get `ScreenBrightnessError`?
Linux often requires a bit of tweaking to get working out of the box. This either means installing a 3rd party program or
granting extra permissions to the user.

These steps are covered in detail on the [Installing On Linux](Installing%20On%20Linux.html) page.


## None of these answered my question
Please [raise an issue](https://github.com/Crozzers/screen_brightness_control/issues/new) on the GitHub repo or email me at [captaincrozzers@gmail.com](mailto:captaincrozzers@gmail.com) with your question.
'''



import ctypes
import logging
import threading
import time
from ctypes import POINTER, WINFUNCTYPE, Structure, WinError, byref, windll
from ctypes.wintypes import (BOOL, BYTE, DWORD, HANDLE, HDC, HMONITOR, LPARAM,
                             RECT, WCHAR)
from typing import List, Optional

import pythoncom
import pywintypes
import win32api
import win32con
import wmi

from . import filter_monitors, get_methods
from .exceptions import EDIDParseError, NoValidDisplayError, format_exc
from .helpers import EDID, BrightnessMethod, __Cache, _monitor_brand_lookup
from .types import DisplayIdentifier, Generator, IntPercentage

__cache__ = __Cache()
logger = logging.getLogger(__name__)


def _wmi_init():
    '''internal function to create and return a wmi instance'''
    # WMI calls don't work in new threads so we have to run this check
    if threading.current_thread() != threading.main_thread():
        pythoncom.CoInitialize()
    return wmi.WMI(namespace='wmi')


def enum_display_devices() -> Generator[win32api.PyDISPLAY_DEVICEType, None, None]:
    '''
    Yields all display devices connected to the computer
    '''
    for monitor_enum in win32api.EnumDisplayMonitors():
        pyhandle = monitor_enum[0]
        monitor_info = win32api.GetMonitorInfo(pyhandle)
        for adaptor_index in range(5):
            try:
                device = win32api.EnumDisplayDevices(
                    monitor_info['Device'], adaptor_index, 1)
            except pywintypes.error:
                logger.debug(
                    f'failed to get display device {monitor_info["Device"]} on adaptor index {adaptor_index}')
            else:
                yield device
                break


def get_display_info() -> List[dict]:
    '''
    Gets information about all connected displays using WMI and win32api

    Example:
        ```python
        import screen_brightness_control as s

        info = s.windows.get_display_info()
        for display in info:
            print(display['name'])
        ```
    '''
    info = __cache__.get('windows_monitors_info_raw')
    if info is None:
        info = []
        # collect all monitor UIDs (derived from DeviceID)
        monitor_uids = {}
        for device in enum_display_devices():
            monitor_uids[device.DeviceID.split('#')[2]] = device

        # gather list of laptop displays to check against later
        wmi = _wmi_init()
        try:
            laptop_displays = [
                i.InstanceName
                for i in wmi.WmiMonitorBrightness()
            ]
        except Exception as e:
            # don't do specific exception classes here because WMI does not play ball with it
            logger.warning(
                f'get_display_info: failed to gather list of laptop displays - {format_exc(e)}')
            laptop_displays = []

        extras, desktop, laptop = [], 0, 0
        uid_keys = list(monitor_uids.keys())
        for monitor in wmi.WmiMonitorDescriptorMethods():
            model, serial, manufacturer, man_id, edid = None, None, None, None, None
            instance_name = monitor.InstanceName.replace(
                '_0', '', 1).split('\\')[2]
            pydevice = monitor_uids[instance_name]

            # get the EDID
            try:
                edid = ''.join(
                    f'{char:02x}' for char in monitor.WmiGetMonitorRawEEdidV1Block(0)[0])
                # we do the EDID parsing ourselves because calling wmi.WmiMonitorID
                # takes too long
                parsed = EDID.parse(edid)
                man_id, manufacturer, model, name, serial = parsed
                if name is None:
                    raise EDIDParseError(
                        'parsed EDID returned invalid display name')
            except EDIDParseError as e:
                edid = None
                logger.warning(
                    f'exception parsing edid str for {monitor.InstanceName} - {format_exc(e)}')
            except Exception as e:
                edid = None
                logger.error(
                    f'failed to get EDID string for {monitor.InstanceName} - {format_exc(e)}')
            finally:
                if edid is None:
                    devid = pydevice.DeviceID.split('#')
                    serial = devid[2]
                    man_id = devid[1][:3]
                    model = devid[1][3:] or 'Generic Monitor'
                    del devid
                    try:
                        man_id, manufacturer = _monitor_brand_lookup(man_id)
                    except TypeError:
                        manufacturer = None

            if (serial, model) != (None, None):
                data = {
                    'name': f'{manufacturer} {model}',
                    'model': model,
                    'serial': serial,
                    'manufacturer': manufacturer,
                    'manufacturer_id': man_id,
                    'edid': edid
                }
                if monitor.InstanceName in laptop_displays:
                    data['index'] = laptop
                    data['method'] = WMI
                    laptop += 1
                else:
                    data['method'] = VCP
                    desktop += 1

                if instance_name in uid_keys:
                    # insert the data into the uid_keys list because
                    # uid_keys has the monitors sorted correctly. This
                    # means we don't have to re-sort the list later
                    uid_keys[uid_keys.index(instance_name)] = data
                else:
                    extras.append(data)

        info = uid_keys + extras
        if desktop:
            # now make sure desktop monitors have the correct index
            count = 0
            for item in info:
                if item['method'] == VCP:
                    item['index'] = count
                    count += 1

        # return info only which has correct data
        info = [i for i in info if isinstance(i, dict)]

        __cache__.store('windows_monitors_info_raw', info)

    return info


class WMI(BrightnessMethod):
    '''
    A collection of screen brightness related methods using the WMI API.
    This class primarily works with laptop displays.
    '''
    @classmethod
    def get_display_info(cls, display: Optional[DisplayIdentifier] = None) -> List[dict]:
        info = [i for i in get_display_info() if i['method'] == cls]
        if display is not None:
            info = filter_monitors(display=display, haystack=info)
        return info

    @classmethod
    def set_brightness(cls, value: IntPercentage, display: Optional[int] = None):
        brightness_method = _wmi_init().WmiMonitorBrightnessMethods()
        if display is not None:
            brightness_method = [brightness_method[display]]

        for method in brightness_method:
            method.WmiSetBrightness(value, 0)

    @classmethod
    def get_brightness(cls, display: Optional[int] = None) -> List[IntPercentage]:
        brightness_method = _wmi_init().WmiMonitorBrightness()
        if display is not None:
            brightness_method = [brightness_method[display]]

        values = [i.CurrentBrightness for i in brightness_method]
        return values


class VCP(BrightnessMethod):
    '''Collection of screen brightness related methods using the DDC/CI commands'''
    _MONITORENUMPROC = WINFUNCTYPE(BOOL, HMONITOR, HDC, POINTER(RECT), LPARAM)

    logger = logger.getChild('VCP')

    class _PHYSICAL_MONITOR(Structure):
        '''internal class, do not call'''
        _fields_ = [('handle', HANDLE),
                    ('description', WCHAR * 128)]

    @classmethod
    def iter_physical_monitors(cls, start: int = 0) -> Generator[ctypes.wintypes.HANDLE, None, None]:
        '''
        A generator to iterate through all physical monitors
        and then close them again afterwards, yielding their handles.
        It is not recommended to use this function unless you are familiar with `ctypes` and `windll`

        Args:
            start: skip the first X handles

        Raises:
            ctypes.WinError: upon failure to enumerate through the monitors
        '''
        def callback(hmonitor, *_):
            monitors.append(HMONITOR(hmonitor))
            return True

        monitors = []
        if not windll.user32.EnumDisplayMonitors(None, None, cls._MONITORENUMPROC(callback), None):
            cls.logger.error('EnumDisplayMonitors failed')
            raise WinError('EnumDisplayMonitors failed')

        # user index keeps track of valid monitors
        user_index = 0
        # monitor index keeps track of valid and pseudo monitors
        monitor_index = 0
        display_devices = list(enum_display_devices())

        wmi = _wmi_init()
        try:
            laptop_displays = [
                i.InstanceName.replace('_0', '').split('\\')[2]
                for i in wmi.WmiMonitorBrightness()
            ]
        except Exception as e:
            cls.logger.warning(
                f'failed to gather list of laptop displays - {format_exc(e)}')
            laptop_displays = []

        for monitor in monitors:
            # Get physical monitor count
            count = DWORD()
            if not windll.dxva2.GetNumberOfPhysicalMonitorsFromHMONITOR(monitor, byref(count)):
                raise WinError(
                    'call to GetNumberOfPhysicalMonitorsFromHMONITOR returned invalid result')
            if count.value > 0:
                # Get physical monitor handles
                physical_array = (cls._PHYSICAL_MONITOR * count.value)()
                if not windll.dxva2.GetPhysicalMonitorsFromHMONITOR(monitor, count.value, physical_array):
                    raise WinError(
                        'call to GetPhysicalMonitorsFromHMONITOR returned invalid result')
                for item in physical_array:
                    # check that the monitor is not a pseudo monitor by
                    # checking it's StateFlags for the
                    # win32con DISPLAY_DEVICE_ATTACHED_TO_DESKTOP flag
                    if display_devices[monitor_index].StateFlags & win32con.DISPLAY_DEVICE_ATTACHED_TO_DESKTOP:
                        # check if monitor is actually a laptop display
                        if display_devices[monitor_index].DeviceID.split('#')[2] not in laptop_displays:
                            if start is None or user_index >= start:
                                yield item.handle
                            # increment user index as a valid monitor was found
                            user_index += 1
                    # increment monitor index
                    monitor_index += 1
                    windll.dxva2.DestroyPhysicalMonitor(item.handle)

    @classmethod
    def get_display_info(cls, display: Optional[DisplayIdentifier] = None) -> List[dict]:
        info = [i for i in get_display_info() if i['method'] == cls]
        if display is not None:
            info = filter_monitors(display=display, haystack=info)
        return info

    @classmethod
    def get_brightness(cls, display: Optional[int] = None, max_tries: int = 50) -> List[IntPercentage]:
        '''
        Args:
            display: the index of the specific display to query.
                If unspecified, all detected displays are queried
            max_tries: the maximum allowed number of attempts to
                read the VCP output from the display

        Returns:
            See `BrightnessMethod.get_brightness`
        '''
        code = BYTE(0x10)
        values = []
        start = display if display is not None else 0
        for index, monitor in enumerate(cls.iter_physical_monitors(start=start), start=start):
            current = __cache__.get(f'vcp_brightness_{index}')
            if current is None:
                cur_out = DWORD()
                handle = HANDLE(monitor)
                for attempt in range(max_tries):
                    if windll.dxva2.GetVCPFeatureAndVCPFeatureReply(handle, code, None, byref(cur_out), None):
                        current = cur_out.value
                        break
                    current = None
                    time.sleep(0.02 if attempt < 20 else 0.1)
                else:
                    cls.logger.error(
                        f'failed to get VCP feature reply for display:{index} after {attempt} tries')

            if current is not None:
                __cache__.store(
                    f'vcp_brightness_{index}', current, expires=0.1)
                values.append(current)

            if display == index:
                # if we have just got the display we wanted then exit here
                # no point iterating through all the other ones
                break

        if 'handle' in locals():
            # make sure final handle is destroyed
            windll.dxva2.DestroyPhysicalMonitor(handle)
        return values

    @classmethod
    def set_brightness(cls, value: IntPercentage, display: Optional[int] = None, max_tries: int = 50):
        '''
        Args:
            display: The specific display you wish to query.
            max_tries: the maximum allowed number of attempts to
                send the VCP input to the display
        '''
        __cache__.expire(startswith='vcp_brightness_')
        code = BYTE(0x10)
        value = DWORD(value)
        start = display if display is not None else 0
        for index, monitor in enumerate(cls.iter_physical_monitors(start=start), start=start):
            if display is None or display == index:
                handle = HANDLE(monitor)
                for attempt in range(max_tries):
                    if windll.dxva2.SetVCPFeature(handle, code, value):
                        break
                    time.sleep(0.02 if attempt < 20 else 0.1)
                else:
                    cls.logger.error(
                        f'failed to set display:{index}->{value} after {attempt} tries')

        if 'handle' in locals():
            # make sure final handle is destroyed
            windll.dxva2.DestroyPhysicalMonitor(handle)


def list_monitors_info(
    method: Optional[str] = None, allow_duplicates: bool = False, unsupported: bool = False
) -> List[dict]:
    '''
    Lists detailed information about all detected displays

    Args:
        method: the method the display can be addressed by. See `screen_brightness_control.get_methods`
            for more info on available methods
        allow_duplicates: whether to filter out duplicate displays (displays with the same EDID) or not
        unsupported: include detected displays that are invalid or unsupported.
            This argument does nothing on Windows
    '''
    # no caching here because get_display_info caches its results
    info = get_display_info()

    all_methods = get_methods(method).values()

    if method is not None:
        info = [i for i in info if i['method'] in all_methods]

    if allow_duplicates:
        return info

    try:
        # use filter_monitors to remove duplicates
        return filter_monitors(haystack=info)
    except NoValidDisplayError:
        return []



import sys

from setuptools import setup

sys.path.insert(0, 'screen_brightness_control')
from _version import __author__, __version__  # noqa: E402

setup(
    name='screen_brightness_control',
    version=__version__,
    url='https://github.com/Crozzers/screen_brightness_control',
    project_urls={
        'Documentation': 'https://crozzers.github.io/screen_brightness_control',
        'Issue Tracker': 'https://github.com/Crozzers/screen_brightness_control/issues'
    },
    license='MIT',
    author=__author__,
    author_email='captaincrozzers@gmail.com',
    packages=['screen_brightness_control'],
    install_requires=open('requirements.txt').readlines(),
    description='A Python tool to control monitor brightness on Windows and Linux',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only'
    ],
    python_requires='>=3.7'
)



import os
import sys
import unittest
from copy import deepcopy
import itertools

sys.path.insert(0, os.path.abspath('./'))
import screen_brightness_control as sbc  # noqa: E402


def get_methods():
    return tuple(sbc.get_methods().values())


def get_method_names():
    return tuple(sbc.get_methods().keys())


class BasicMethodTest(object):
    def assertBrightnessEqual(self, a, b):
        try:
            self.assertEqual(a, b)
        except AssertionError as e:
            if os.name == 'nt' and self.method == sbc.windows.WMI:
                # some laptop displays will only have a set number of levels
                # the brightness can be set to so check the value is at least close.
                # this allows for 16ish brightness levels
                self.assertTrue(abs(a - b) < 7)
            else:
                raise e

    def test_get_display_info(self):
        displays = self.method.get_display_info()
        self.assertIsInstance(displays, list)
        for display in displays:
            self.assertIsInstance(display, dict)
            self.assertEqual(display['method'], self.method)

            for key in ('name', 'model', 'serial', 'manufacturer', 'manufacturer_id', 'edid'):
                self.assertIn(key, display)
                self.assertIsInstance(display[key], (str, type(None)))

            self.assertIn('index', display)
            self.assertIsInstance(display['index'], int)

    def test_get_brightness(self):
        displays = self.method.get_display_info()
        if displays:  # if there are any displays this method can address
            brightness = self.method.get_brightness()
            self.assertIsInstance(brightness, list)
            self.assertEqual(len(displays), len(brightness))

            for value in brightness:
                self.assertIsInstance(value, int)
                self.assertTrue(0 <= value <= 100)

            single_brightness = self.method.get_brightness(display=0)
            self.assertIsInstance(single_brightness, list)
            self.assertEqual(len(single_brightness), 1)

    def test_set_brightness(self):
        if self.method.get_display_info():  # if there are any displays this method can address
            for i in (0, 10, 25, 45, 50, 60, 71, 80, 99, 100):
                self.assertIsNone(self.method.set_brightness(i))

                for value in self.method.get_brightness():
                    self.assertIsInstance(value, int)
                    self.assertBrightnessEqual(value, i)


class FakeMethodTest():
    @classmethod
    def generate_fake_displays(cls):
        fakes = []
        for i in range(3):
            fakes.append(
                {
                    'name': 'FakeDisplay ABC123',
                    'model': 'ABC123',
                    'manufacturer': 'FakeDisplay',
                    'manufacturer_id': None,
                    'serial': f'FakeSerialDEF456_{i}',
                    'index': i,
                    'edid': None,
                    'method': cls
                }
            )
        fakes.append(deepcopy(fakes[-1]))
        fakes[-1]['unsupported'] = True
        return fakes

    def __enter__(self):
        if not hasattr(sbc, '_old_methods'):
            sbc._old_os_methods = sbc._OS_METHODS

        # change the sbc get_methods function output
        sbc._OS_METHODS = FAKE_METHODS
        # on windows we also need to override the module level `get_display_info`
        if os.name == 'nt':
            if not hasattr(sbc._OS_MODULE, '_old_get_display_info'):
                sbc._OS_MODULE._old_get_display_info = sbc._OS_MODULE.get_display_info

            def get_display_info():
                displays = itertools.chain.from_iterable(i.generate_fake_displays() for i in FAKE_METHODS)
                if os.name == 'nt':
                    return self._remove_unsupported_displays(displays)
                return displays

            sbc._OS_MODULE.get_display_info = get_display_info

    def __exit__(self, *args):
        # change the sbc get_methods function output
        sbc._OS_METHODS = sbc._old_os_methods
        # on windows we also need to override the module level `get_display_info`
        if os.name == 'nt':
            sbc._OS_MODULE.get_display_info = sbc._OS_MODULE._old_get_display_info

    cached_display_info = []
    cached_brightness = {}

    @classmethod
    def _gdi(cls):
        if not cls.cached_display_info:
            for method in sbc._old_os_methods:
                try:
                    for d in method.get_display_info():
                        d['method'] = cls
                        cls.cached_display_info.append(d)
                except Exception:
                    pass

            if not cls.cached_display_info:
                # no display info available. Lets make some up
                cls.cached_display_info = cls.generate_fake_displays()

            cls.cached_display_info = sbc.filter_monitors(haystack=cls.cached_display_info)

        return cls.cached_display_info

    @staticmethod
    def _remove_unsupported_displays(displays):
        result = []
        for display in displays:
            if display.get('unsupported'):
                continue
            if 'unsupported' in display:
                display.pop('unsupported')
            result.append(display)
        return result

    @classmethod
    def get_display_info(cls, display=None):
        display_info = cls._remove_unsupported_displays(cls._gdi())
        return sbc.filter_monitors(display=display, haystack=display_info)

    @classmethod
    def get_brightness(cls, display=None):
        all_displays = cls.get_display_info(display)

        results = []
        for display in all_displays:
            cache_ident = '%s-%s-%s' % (display['name'], display['model'], display['serial'])
            if cache_ident not in cls.cached_brightness:
                cls.cached_brightness[cache_ident] = 100

            results.append(cls.cached_brightness[cache_ident])

        return results

    @classmethod
    def set_brightness(cls, value, display=None):
        all_displays = cls.get_display_info(display)

        for display in all_displays:
            cache_ident = '%s-%s-%s' % (display['name'], display['model'], display['serial'])
            cls.cached_brightness[cache_ident] = value


class FakeMethodTest2(FakeMethodTest):
    @classmethod
    def generate_fake_displays(cls, *args, **kwargs):
        displays = deepcopy(FakeMethodTest.generate_fake_displays(*args, **kwargs))
        for display in displays:
            display['method'] = cls
        return cls._remove_unsupported_displays(displays)

    @classmethod
    def get_display_info(cls, *args, **kwargs):
        displays = deepcopy(FakeMethodTest.get_display_info(*args, **kwargs))
        for display in displays:
            display['method'] = cls
        return cls._remove_unsupported_displays(displays)


class FakeMethodTest3(FakeMethodTest):
    @classmethod
    def generate_fake_displays(cls, *args, **kwargs):
        displays = deepcopy(FakeMethodTest.generate_fake_displays(*args, **kwargs))
        for display in displays:
            display['method'] = cls
        return cls._remove_unsupported_displays(displays)

    @classmethod
    def get_display_info(cls, *args, **kwargs):
        displays = deepcopy(FakeMethodTest.get_display_info(*args, **kwargs))
        for display in displays:
            display['method'] = cls
        return cls._remove_unsupported_displays(displays)


class FakeMethodTest4(FakeMethodTest):
    # this FakeMethodTest is slightly different from 1-3. Those all return
    # duplicate displays with the method changed. This one returns no displays.
    # See #21
    @classmethod
    def generate_fake_displays(cls, *args, **kwargs):
        return []

    @classmethod
    def get_display_info(cls, *args, **kwargs):
        return []


FAKE_METHODS = [FakeMethodTest, FakeMethodTest2, FakeMethodTest3, FakeMethodTest4]


class TestCase(unittest.TestCase):
    def setUp(self):
        if not TEST_FAST:
            # only set brightness to 100 pre test if testing in slow mode
            sbc.set_brightness(100, verbose_error=True)
        else:
            FakeMethodTest().__enter__()

    def tearDown(self):
        if not TEST_FAST:
            # only set brightness to 100 post test if testing in slow mode
            sbc.set_brightness(100, verbose_error=True)
        else:
            FakeMethodTest().__exit__()

    def assertBrightnessEqual(self, a, b, display):
        try:
            self.assertEqual(a, b)
        except AssertionError as e:
            if os.name == 'nt':
                if sbc.list_monitors_info()[display]['method'] == sbc.windows.WMI:
                    # some laptop displays will only have a set number of levels
                    # the brightness can be set to so check the value is at least close.
                    # this allows for 16ish brightness levels
                    self.assertTrue(abs(a - b) < 7)
                else:
                    raise e
            else:  # linux
                if a < 2 and b < 2:
                    # on linux if you set the brightness to 0 it actually sets it to 1
                    pass
                else:
                    raise e

    def assertBrightnessValid(self, brightness, target_length=None):
        self.assertIsInstance(brightness, list)
        for value in brightness:
            self.assertIsInstance(value, int)
            self.assertTrue(0 <= value <= 100)

        if target_length is not None:
            self.assertTrue(len(brightness) == target_length)


TEST_FAST = False
